from typing import Callable
import numpy


def calc_iint(
        beam_polarization: float, flipper_efficiency: float, f_nucl, f_m_perp, matrix_u, func_extinction: Callable = None,
        flag_beam_polarization: bool = False, flag_flipper_efficiency: bool = False,
        flag_f_nucl: bool = False, flag_f_m_perp: bool = False,
        dict_in_out: dict = None, flag_use_precalculated_data: bool = False):
    """Calculate integrated intensities.

    It is supposed that crystal is not rotate during the calculations 
    (orientation matrix is the same for all reflections)
    """
    if dict_in_out is None:
        flag_dict = False
        dict_in_out_keys = []
    else:
        flag_dict = True
        dict_in_out_keys = dict_in_out.keys()

    p_u = beam_polarization
    p_d = beam_polarization*(2*flipper_efficiency-1)
    flag_f_plus_sq = flag_f_nucl or flag_f_m_perp
    flag_f_minus_sq = flag_f_nucl or flag_f_m_perp
    flag_f_m_perp_xy_sq = flag_f_m_perp

    f_n = numpy.atleast_1d(f_nucl)
    f_m_perp = numpy.atleast_1d(f_m_perp)
    f_n_sq = numpy.square(numpy.abs(f_n))

    f_m_perp_x = f_m_perp[0]*matrix_u[0] + f_m_perp[1]*matrix_u[1] + f_m_perp[2]*matrix_u[2]
    f_m_perp_y = f_m_perp[0]*matrix_u[3] + f_m_perp[1]*matrix_u[4] + f_m_perp[2]*matrix_u[5]
    f_m_perp_z = f_m_perp[0]*matrix_u[6] + f_m_perp[1]*matrix_u[7] + f_m_perp[2]*matrix_u[8]
    
    f_m_perp_x_sq = numpy.square(numpy.abs(f_m_perp_x))
    f_m_perp_y_sq = numpy.square(numpy.abs(f_m_perp_y))
    f_m_perp_z_sq = numpy.square(numpy.abs(f_m_perp_z))

    f_n_f_m_perp_z = 2.*(f_n.real * f_m_perp_z.real + f_n.imag * f_m_perp_z.imag)

    f_plus_sq = f_n_sq + f_m_perp_z_sq + f_n_f_m_perp_z
    f_minus_sq = f_n_sq + f_m_perp_z_sq - f_n_f_m_perp_z
    f_m_perp_xy_sq = f_m_perp_x_sq + f_m_perp_y_sq

    if func_extinction is None:
        dder_y_plus, dder_y_minus, dder_y_m_perp_xy = {}, {}, {}
        y_plus = numpy.ones_like(f_plus_sq)
        y_minus = numpy.ones_like(f_minus_sq)
        y_m_perp_xy = numpy.ones_like(f_m_perp_xy_sq)
    else:
        y_plus, dder_y_plus = func_extinction(f_plus_sq, flag_f_sq=flag_f_plus_sq)
        y_minus, dder_y_minus = func_extinction(f_minus_sq, flag_f_sq=flag_f_minus_sq)
        y_m_perp_xy, dder_y_m_perp_xy = func_extinction(f_m_perp_xy_sq, flag_f_sq=flag_f_m_perp_xy_sq)

    chiral_term = 2.*(f_m_perp_x.imag * f_m_perp_y.real - f_m_perp_x.real * f_m_perp_y.imag)
    if flag_dict:
        dict_in_out["chiral_term"] = chiral_term

    iint_plus = 0.5*((1.+p_u)*y_plus*f_plus_sq +
                     (1.-p_u)*y_minus*f_minus_sq) + \
                    y_m_perp_xy * f_m_perp_xy_sq + \
                    p_u * chiral_term

    iint_minus = 0.5*((1.-p_d)*y_plus*f_plus_sq +
                     (1.+p_d)*y_minus*f_minus_sq) + \
                    y_m_perp_xy * f_m_perp_xy_sq - \
                    p_d * chiral_term

    if flag_dict:
        dict_in_out["iint_plus"] = iint_plus
        dict_in_out["iint_minus"] = iint_minus
        dict_in_out["y_plus"] = y_plus
        dict_in_out["f_plus_sq"] = f_plus_sq
        dict_in_out["y_minus"] = y_minus
        dict_in_out["f_minus_sq"] = f_minus_sq
        dict_in_out["y_m_perp_xy"] = y_m_perp_xy
        dict_in_out["f_m_perp_xy_sq"] = f_m_perp_xy_sq

    dder_plus = {}
    dder_minus = {}
    if flag_beam_polarization:
        dder_plus["beam_polarization"] = 0.5*(y_plus*f_plus_sq - y_minus*f_minus_sq + 2.*chiral_term) * \
            numpy.ones_like(beam_polarization)
        dder_minus["beam_polarization"] = 0.5*(-y_plus*f_plus_sq + y_minus*f_minus_sq - 2.*chiral_term) * \
            numpy.ones_like(beam_polarization)*(2.*flipper_efficiency-1.)

    if flag_flipper_efficiency:
        dder_minus["flipper_efficiency"] = beam_polarization*(-y_plus*f_plus_sq + y_minus*f_minus_sq - 2.*chiral_term) * \
            numpy.ones_like(flipper_efficiency)
    
    if flag_f_nucl:
        f_plus_sq_f_n_real = 2. * (f_n.real + f_m_perp_z.real) * numpy.ones_like(f_n.real)
        f_plus_sq_f_n_imag = 2. * (f_n.imag + f_m_perp_z.imag) * numpy.ones_like(f_n.imag)
        f_minus_sq_f_n_real = 2. * (f_n.real - f_m_perp_z.real)* numpy.ones_like(f_n.real)
        f_minus_sq_f_n_imag = 2. * (f_n.imag - f_m_perp_z.imag) * numpy.ones_like(f_n.imag)

        y_plus_f_n_real, y_minus_f_n_real = 0, 0 
        y_plus_f_n_imag, y_minus_f_n_imag = 0, 0 
        if "f_sq" in dder_y_plus.keys(): 
            y_plus_f_n_real = dder_y_plus["f_sq"]*f_plus_sq_f_n_real
            y_plus_f_n_imag = dder_y_plus["f_sq"]*f_plus_sq_f_n_imag
        if "f_sq" in dder_y_minus.keys(): 
            y_minus_f_n_real = dder_y_minus["f_sq"]*f_minus_sq_f_n_real
            y_minus_f_n_imag = dder_y_minus["f_sq"]*f_minus_sq_f_n_imag

        dder_plus["f_nucl_real"] =  0.5*(
            (1.+p_u)*(y_plus*f_plus_sq_f_n_real+y_plus_f_n_real*f_plus_sq) +
            (1.-p_u)*(y_minus*f_minus_sq_f_n_real+y_minus_f_n_real*f_minus_sq))
        
        dder_plus["f_nucl_imag"] =  0.5*(
            (1.+p_u)*(y_plus*f_plus_sq_f_n_imag+y_plus_f_n_imag*f_plus_sq) +
            (1.-p_u)*(y_minus*f_minus_sq_f_n_imag+y_minus_f_n_imag*f_minus_sq))

        dder_minus["f_nucl_real"] =  0.5*(
            (1.-p_d)*(y_plus*f_plus_sq_f_n_real+y_plus_f_n_real*f_plus_sq) +
            (1.+p_d)*(y_minus*f_minus_sq_f_n_real+y_minus_f_n_real*f_minus_sq))
        
        dder_minus["f_nucl_imag"] =  0.5*(
            (1.-p_d)*(y_plus*f_plus_sq_f_n_imag+y_plus_f_n_imag*f_plus_sq) +
            (1.+p_d)*(y_minus*f_minus_sq_f_n_imag+y_minus_f_n_imag*f_minus_sq))


    if flag_f_m_perp:
        f_plus_sq_f_m_perp_z_real = 2. * (f_n.real + f_m_perp_z.real) * numpy.ones_like(f_m_perp_z.real)
        f_plus_sq_f_m_perp_z_imag = 2. * (f_n.imag + f_m_perp_z.imag) * numpy.ones_like(f_m_perp_z.imag)
        f_minus_sq_f_m_perp_z_real = -2. * (f_n.real - f_m_perp_z.real) * numpy.ones_like(f_m_perp_z.real)
        f_minus_sq_f_m_perp_z_imag = -2. * (f_n.imag - f_m_perp_z.imag) * numpy.ones_like(f_m_perp_z.imag)
        f_m_perp_xy_sq_f_m_perp_x_real = 2 * f_m_perp_x.real * numpy.ones_like(f_m_perp_x.real)
        f_m_perp_xy_sq_f_m_perp_x_imag = 2 * f_m_perp_x.imag * numpy.ones_like(f_m_perp_x.imag)
        f_m_perp_xy_sq_f_m_perp_y_real = 2 * f_m_perp_y.real * numpy.ones_like(f_m_perp_y.real)
        f_m_perp_xy_sq_f_m_perp_y_imag = 2 * f_m_perp_y.imag * numpy.ones_like(f_m_perp_y.imag)
        chiral_term_f_m_perp_x_real = -2 * f_m_perp_y.imag * numpy.ones_like(f_m_perp_x.real)
        chiral_term_f_m_perp_x_imag = 2 * f_m_perp_y.real * numpy.ones_like(f_m_perp_x.imag)
        chiral_term_f_m_perp_y_real = 2 * f_m_perp_x.imag * numpy.ones_like(f_m_perp_y.real)
        chiral_term_f_m_perp_y_imag = -2 * f_m_perp_x.real * numpy.ones_like(f_m_perp_y.imag)

        y_plus_f_m_perp_z_real, y_plus_f_m_perp_z_imag = 0, 0 
        y_minus_f_m_perp_z_real, y_minus_f_m_perp_z_imag = 0, 0 
        y_m_perp_xy_f_m_perp_x_real, y_m_perp_xy_f_m_perp_x_imag = 0, 0 
        y_m_perp_xy_f_m_perp_y_real, y_m_perp_xy_f_m_perp_y_imag = 0, 0 
        if "f_sq" in dder_y_plus.keys(): 
            y_plus_f_m_perp_z_real = dder_y_plus["f_sq"]*f_plus_sq_f_m_perp_z_real
            y_plus_f_m_perp_z_imag = dder_y_plus["f_sq"]*f_plus_sq_f_m_perp_z_imag
        if "f_sq" in dder_y_minus.keys(): 
            y_minus_f_m_perp_z_real = dder_y_minus["f_sq"]*f_minus_sq_f_m_perp_z_real
            y_minus_f_m_perp_z_imag = dder_y_minus["f_sq"]*f_minus_sq_f_m_perp_z_imag
        if "f_sq" in dder_y_m_perp_xy.keys(): 
            y_m_perp_xy_f_m_perp_x_real = dder_y_m_perp_xy["f_sq"]*f_m_perp_xy_sq_f_m_perp_x_real
            y_m_perp_xy_f_m_perp_x_imag = dder_y_m_perp_xy["f_sq"]*f_m_perp_xy_sq_f_m_perp_x_imag
            y_m_perp_xy_f_m_perp_y_real = dder_y_m_perp_xy["f_sq"]*f_m_perp_xy_sq_f_m_perp_y_real
            y_m_perp_xy_f_m_perp_y_imag = dder_y_m_perp_xy["f_sq"]*f_m_perp_xy_sq_f_m_perp_y_imag

        dder_plus_f_m_perp_x_real = \
            y_m_perp_xy_f_m_perp_x_real * f_m_perp_xy_sq + \
            y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_x_real +\
            p_u * chiral_term_f_m_perp_x_real
        dder_plus_f_m_perp_x_imag = \
            y_m_perp_xy_f_m_perp_x_imag * f_m_perp_xy_sq + \
            y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_x_imag +\
            p_u * chiral_term_f_m_perp_x_imag
        dder_plus_f_m_perp_y_real = \
            y_m_perp_xy_f_m_perp_y_real * f_m_perp_xy_sq + \
            y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_y_real +\
            p_u * chiral_term_f_m_perp_y_real
        dder_plus_f_m_perp_y_imag = \
            y_m_perp_xy_f_m_perp_y_imag * f_m_perp_xy_sq + \
            y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_y_imag +\
            p_u * chiral_term_f_m_perp_y_imag
        dder_plus_f_m_perp_z_real =  0.5*(
            (1.+p_u)*(y_plus*f_plus_sq_f_m_perp_z_real+y_plus_f_m_perp_z_real*f_plus_sq) +
            (1.-p_u)*(y_minus*f_minus_sq_f_m_perp_z_real+y_minus_f_m_perp_z_real*f_minus_sq))
        dder_plus_f_m_perp_z_imag =  0.5*(
            (1.+p_u)*(y_plus*f_plus_sq_f_m_perp_z_imag+y_plus_f_m_perp_z_imag*f_plus_sq) +
            (1.-p_u)*(y_minus*f_minus_sq_f_m_perp_z_imag+y_minus_f_m_perp_z_imag*f_minus_sq))

        dder_plus_f_m_perp_1_real = \
            dder_plus_f_m_perp_x_real*matrix_u[0] + \
            dder_plus_f_m_perp_y_real*matrix_u[3] + \
            dder_plus_f_m_perp_z_real*matrix_u[6]
        dder_plus_f_m_perp_1_imag = \
            dder_plus_f_m_perp_x_imag*matrix_u[0] + \
            dder_plus_f_m_perp_y_imag*matrix_u[3] + \
            dder_plus_f_m_perp_z_imag*matrix_u[6]
        dder_plus_f_m_perp_2_real = \
            dder_plus_f_m_perp_x_real*matrix_u[1] + \
            dder_plus_f_m_perp_y_real*matrix_u[4] + \
            dder_plus_f_m_perp_z_real*matrix_u[7]
        dder_plus_f_m_perp_2_imag = \
            dder_plus_f_m_perp_x_imag*matrix_u[1] + \
            dder_plus_f_m_perp_y_imag*matrix_u[4] + \
            dder_plus_f_m_perp_z_imag*matrix_u[7]
        dder_plus_f_m_perp_3_real = \
            dder_plus_f_m_perp_x_real*matrix_u[2] + \
            dder_plus_f_m_perp_y_real*matrix_u[5] + \
            dder_plus_f_m_perp_z_real*matrix_u[8]
        dder_plus_f_m_perp_3_imag = \
            dder_plus_f_m_perp_x_imag*matrix_u[2] + \
            dder_plus_f_m_perp_y_imag*matrix_u[5] + \
            dder_plus_f_m_perp_z_imag*matrix_u[8]

        dder_plus["f_m_perp_real"] = numpy.stack([
            dder_plus_f_m_perp_1_real, dder_plus_f_m_perp_2_real, dder_plus_f_m_perp_3_real],
            axis=0) 
        dder_plus["f_m_perp_imag"] = numpy.stack([
            dder_plus_f_m_perp_1_imag, dder_plus_f_m_perp_2_imag, dder_plus_f_m_perp_3_imag],
            axis=0)


        dder_minus_f_m_perp_x_real = \
            y_m_perp_xy_f_m_perp_x_real * f_m_perp_xy_sq + \
            y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_x_real -\
            p_d * chiral_term_f_m_perp_x_real
        dder_minus_f_m_perp_x_imag = \
            y_m_perp_xy_f_m_perp_x_imag * f_m_perp_xy_sq + \
            y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_x_imag -\
            p_d * chiral_term_f_m_perp_x_imag
        dder_minus_f_m_perp_y_real = \
            y_m_perp_xy_f_m_perp_y_real * f_m_perp_xy_sq + \
            y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_y_real -\
            p_d * chiral_term_f_m_perp_y_real
        dder_minus_f_m_perp_y_imag = \
            y_m_perp_xy_f_m_perp_y_imag * f_m_perp_xy_sq + \
            y_m_perp_xy * f_m_perp_xy_sq_f_m_perp_y_imag -\
            p_d * chiral_term_f_m_perp_y_imag
        dder_minus_f_m_perp_z_real =  0.5*(
            (1.-p_d)*(y_plus*f_plus_sq_f_m_perp_z_real+y_plus_f_m_perp_z_real*f_plus_sq) +
            (1.+p_d)*(y_minus*f_minus_sq_f_m_perp_z_real+y_minus_f_m_perp_z_real*f_minus_sq))
        dder_minus_f_m_perp_z_imag =  0.5*(
            (1.-p_d)*(y_plus*f_plus_sq_f_m_perp_z_imag+y_plus_f_m_perp_z_imag*f_plus_sq) +
            (1.+p_d)*(y_minus*f_minus_sq_f_m_perp_z_imag+y_minus_f_m_perp_z_imag*f_minus_sq))

        dder_minus_f_m_perp_1_real = \
            dder_minus_f_m_perp_x_real*matrix_u[0] + \
            dder_minus_f_m_perp_y_real*matrix_u[3] + \
            dder_minus_f_m_perp_z_real*matrix_u[6]
        dder_minus_f_m_perp_1_imag = \
            dder_minus_f_m_perp_x_imag*matrix_u[0] + \
            dder_minus_f_m_perp_y_imag*matrix_u[3] + \
            dder_minus_f_m_perp_z_imag*matrix_u[6]
        dder_minus_f_m_perp_2_real = \
            dder_minus_f_m_perp_x_real*matrix_u[1] + \
            dder_minus_f_m_perp_y_real*matrix_u[4] + \
            dder_minus_f_m_perp_z_real*matrix_u[7]
        dder_minus_f_m_perp_2_imag = \
            dder_minus_f_m_perp_x_imag*matrix_u[1] + \
            dder_minus_f_m_perp_y_imag*matrix_u[4] + \
            dder_minus_f_m_perp_z_imag*matrix_u[7]
        dder_minus_f_m_perp_3_real = \
            dder_minus_f_m_perp_x_real*matrix_u[2] + \
            dder_minus_f_m_perp_y_real*matrix_u[5] + \
            dder_minus_f_m_perp_z_real*matrix_u[8]
        dder_minus_f_m_perp_3_imag = \
            dder_minus_f_m_perp_x_imag*matrix_u[2] + \
            dder_minus_f_m_perp_y_imag*matrix_u[5] + \
            dder_minus_f_m_perp_z_imag*matrix_u[8]

        dder_minus["f_m_perp_real"] = numpy.stack([
            dder_minus_f_m_perp_1_real, dder_minus_f_m_perp_2_real, dder_minus_f_m_perp_3_real], 
            axis=0) 
        dder_minus["f_m_perp_imag"] = numpy.stack([
            dder_minus_f_m_perp_1_imag, dder_minus_f_m_perp_2_imag, dder_minus_f_m_perp_3_imag], 
            axis=0)

    extinction_keys = dder_y_plus.keys()
    if len(extinction_keys) != 0:
        for key in extinction_keys:
            if key == "f_sq":
                pass
            else:
                dder_plus[key] = \
                    0.5*((1.+p_u)*f_plus_sq*dder_y_plus[key] +
                         (1.-p_u)*f_minus_sq*dder_y_minus[key]) + \
                    f_m_perp_xy_sq * dder_y_m_perp_xy[key]
                dder_minus[key] = \
                    0.5*((1.-p_d)*f_plus_sq*dder_y_plus[key] +
                         (1.+p_d)*f_minus_sq*dder_y_minus[key]) + \
                    f_m_perp_xy_sq * dder_y_m_perp_xy[key]
    return iint_plus, iint_minus, dder_plus, dder_minus


def calc_flip_ratio_by_iint(
        iint_plus, iint_minus, c_lambda2: float = None, iint_2hkl = None,
        flag_iint_plus: bool = False, flag_iint_minus: bool = False, 
        flag_c_lambda2: bool = False, flag_iint_2hkl: bool = False):
    """Calculate flip ratio by given integrated intensities.
    """
    dder = {}
    cond = (c_lambda2 is None) or (iint_2hkl is None)
    if cond:
        signal_plus = iint_plus 
        signal_minus = iint_minus 
    else:
        signal_plus = iint_plus + c_lambda2*iint_2hkl
        signal_minus = iint_minus + c_lambda2*iint_2hkl

    flip_ratio = signal_plus/signal_minus

    if flag_iint_plus:
        dder["iint_plus"] = numpy.ones_like(iint_plus) / signal_minus
    if flag_iint_minus:
        dder["iint_minus"] = -1. * flip_ratio * numpy.ones_like(iint_minus)/ signal_minus
    if flag_c_lambda2:
        dder["c_lambda2"] = numpy.ones_like(c_lambda2)*iint_2hkl / signal_minus - \
            flip_ratio * numpy.ones_like(c_lambda2)*iint_2hkl/ signal_minus
    if flag_iint_2hkl:
        dder["iint_2hkl"] = numpy.ones_like(iint_2hkl)*c_lambda2 / signal_minus - \
            flip_ratio * numpy.ones_like(iint_2hkl)*c_lambda2/ signal_minus
    return flip_ratio, dder


def calc_asymmetry_by_iint(
        iint_plus, iint_minus, c_lambda2: float = None, iint_2hkl = None,
        flag_iint_plus: bool = False, flag_iint_minus: bool = False, 
        flag_c_lambda2: bool = False, flag_iint_2hkl: bool = False):
    """Calculate asymmetry (I^+ - I^-)/(I^+ + I^-).
    """
    dder = {}
    cond = (c_lambda2 is None) or (iint_2hkl is None)
    if cond:
        contamination = 0.
    else:
        contamination = 2.*c_lambda2*iint_2hkl

    denom = (iint_plus + iint_minus + contamination)
    asymmetry = (iint_plus - iint_minus) / denom

    if flag_iint_plus:
        dder["iint_plus"] = (1. / denom - (iint_plus - iint_minus) / numpy.square(denom)) * \
            numpy.ones_like(iint_plus)
    if flag_iint_minus:
        dder["iint_minus"] = (-1. / denom - (iint_plus - iint_minus) / numpy.square(denom)) * \
            numpy.ones_like(iint_minus)
    if flag_c_lambda2:
        dder["c_lambda2"] = 2.*iint_2hkl*( - (iint_plus - iint_minus) / numpy.square(denom)) * \
            numpy.ones_like(c_lambda2)
    if flag_iint_2hkl:
        dder["iint_2hkl"] = 2.*c_lambda2*( - (iint_plus - iint_minus) / numpy.square(denom)) * \
            numpy.ones_like(iint_2hkl)
    return asymmetry, dder


def calc_intensities_by_structure_factors(
        beam_polarization: float, flipper_efficiency: float, f_nucl, f_m_perp, matrix_u, func_extinction: Callable = None,
        c_lambda2: float=None, f_nucl_2hkl=None, f_m_perp_2hkl=None,
        flag_beam_polarization: bool = False, flag_flipper_efficiency: bool = False,
        flag_f_nucl: bool = False, flag_f_m_perp: bool = False,
        flag_c_lambda2: bool = False,
        flag_f_nucl_2hkl: bool = False, flag_f_m_perp_2hkl: bool = False,
        dict_in_out: dict = None, flag_use_precalculated_data: bool = False):
    """Calculate flip ratio by given structure factors
    """
    if dict_in_out is None:
        flag_dict = False
        dict_in_out_keys = []
    else:
        flag_dict = True
        dict_in_out_keys = dict_in_out.keys()
    
    iint_plus, iint_minus, dder_plus_hkl, dder_minus_hkl = calc_iint(
        beam_polarization, flipper_efficiency, f_nucl, f_m_perp, matrix_u, func_extinction=func_extinction,
        flag_beam_polarization=flag_beam_polarization, flag_flipper_efficiency=flag_flipper_efficiency,
        flag_f_nucl=flag_f_nucl, flag_f_m_perp=flag_f_m_perp,
        dict_in_out=dict_in_out, flag_use_precalculated_data=flag_use_precalculated_data)

    cond = (c_lambda2 is None) or (f_nucl_2hkl is None) or (f_m_perp_2hkl is None)
    if cond:
        c_lambda2 = None
        iint_2hkl = None
        flag_iint_2hkl = False
        dder_2hkl = {}
    else:
        iint_plus_2hkl, iint_minus_2hkl, dder_plus_2hkl, dder_minus_2hkl = calc_iint(
            0, 0, f_nucl_2hkl, f_m_perp_2hkl, matrix_u, func_extinction=func_extinction,
            flag_beam_polarization=False, flag_flipper_efficiency=False,
            flag_f_nucl=flag_f_nucl_2hkl, flag_f_m_perp=flag_f_m_perp_2hkl,
            dict_in_out=None, flag_use_precalculated_data=False)
        iint_2hkl = iint_plus_2hkl
        dder_2hkl = dder_plus_2hkl
        flag_iint_2hkl = len(dder_2hkl.keys()) > 0
        iint_plus += c_lambda2*iint_2hkl
        iint_minus += c_lambda2*iint_2hkl

        if flag_f_nucl_2hkl:
            dder_2hkl["f_nucl_2hkl_real"] = dder_2hkl.pop("f_nucl_real")
            dder_2hkl["f_nucl_2hkl_imag"] = dder_2hkl.pop("f_nucl_imag")
        if flag_f_m_perp_2hkl:
            dder_2hkl["f_m_perp_2hkl_real"] = dder_2hkl.pop("f_m_perp_real")
            dder_2hkl["f_m_perp_2hkl_imag"] = dder_2hkl.pop("f_m_perp_imag")

    dder_plus = {}
    dder_minus = {}
    keys_plus = dder_plus_hkl.keys()
    keys_minus = dder_minus_hkl.keys()
    keys_2h = dder_2hkl.keys()
    set_key = set(keys_plus) | set(keys_minus) | set(keys_2h)
    for key in set_key:
        flag_1 = key in keys_plus
        flag_2 = key in keys_minus
        flag_3 = key in keys_2h
        if flag_1:
            term_1 = dder_plus_hkl[key]
        if flag_2:
            term_2 = dder_minus_hkl[key]
        if flag_3:
            term_3 = dder_2hkl[key]

        if flag_1 and flag_3:
            dder_plus[key] = term_1 + c_lambda2 * term_3
        elif flag_1:
            dder_plus[key] = term_1 
        elif flag_3:
            dder_plus[key] = c_lambda2 * term_3
        
        if flag_2 and flag_3:
            dder_minus[key] = term_2 + c_lambda2 * term_3
        elif flag_2:
            dder_minus[key] = term_2
        elif flag_3:
            dder_minus[key] = c_lambda2 * term_3

    if flag_c_lambda2:
        dder_plus["c_lambda2"] = iint_2hkl
        dder_minus["c_lambda2"] = iint_2hkl
    return iint_plus, iint_minus, dder_plus, dder_minus


def calc_flip_ratio_by_structure_factors(
        beam_polarization: float, flipper_efficiency: float, f_nucl, f_m_perp, matrix_u, func_extinction: Callable = None,
        c_lambda2: float=None, f_nucl_2hkl=None, f_m_perp_2hkl=None,
        flag_beam_polarization: bool = False, flag_flipper_efficiency: bool = False,
        flag_f_nucl: bool = False, flag_f_m_perp: bool = False,
        flag_c_lambda2: bool = False,
        flag_f_nucl_2hkl: bool = False, flag_f_m_perp_2hkl: bool = False,
        dict_in_out: dict = None, flag_use_precalculated_data: bool = False):
    """Calculate flip ratio by given structure factors
    """
    if dict_in_out is None:
        flag_dict = False
        dict_in_out_keys = []
    else:
        flag_dict = True
        dict_in_out_keys = dict_in_out.keys()

    iint_plus, iint_minus, dder_plus, dder_minus = calc_intensities_by_structure_factors(
        beam_polarization, flipper_efficiency, f_nucl, f_m_perp, matrix_u, func_extinction=func_extinction,
        c_lambda2=c_lambda2, f_nucl_2hkl=f_nucl_2hkl, f_m_perp_2hkl=f_m_perp_2hkl,
        flag_beam_polarization=flag_beam_polarization, flag_flipper_efficiency=flag_flipper_efficiency,
        flag_f_nucl=flag_f_nucl, flag_f_m_perp=flag_f_m_perp,
        flag_c_lambda2=flag_c_lambda2,
        flag_f_nucl_2hkl=flag_f_nucl_2hkl, flag_f_m_perp_2hkl=flag_f_m_perp_2hkl,
        dict_in_out=dict_in_out, flag_use_precalculated_data=flag_use_precalculated_data)

    flip_ration = iint_plus/iint_minus

    dder = {}
    keys_plus = dder_plus.keys()
    keys_minus = dder_minus.keys()
    set_key = set(keys_plus) | set(keys_minus)
    for key in set_key:
        flag_1 = key in keys_plus
        flag_2 = key in keys_minus
        if flag_1 and flag_2:
            dder[key] = dder_plus[key]/iint_minus - flip_ration*dder_minus[key]/iint_minus
        elif flag_1:
            dder[key] = dder_plus[key]/iint_minus
        elif flag_2:
            dder[key] = - flip_ration*dder_minus[key]/iint_minus
    return flip_ration, dder





import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.C_item_loop_classes.cl_1_pd_meas import PdMeasL

na = numpy.newaxis

class PdBackground(ItemN):
    """Background point in powder 1d experiment.
    """
    ATTR_MANDATORY_NAMES = ("ttheta", "intensity")
    ATTR_MANDATORY_TYPES = (float, float)
    ATTR_MANDATORY_CIF = ("2theta", "intensity")

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("intensity", )
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "pd_background"

    def __init__(self, **kwargs) -> NoReturn:
        super(PdBackground, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"ttheta": 0}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class PdBackgroundL(LoopN):
    """Background points in powder 1d experiment.
    """
    ITEM_CLASS = PdBackground
    ATTR_INDEX = "ttheta"
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(PdBackgroundL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name
   
    def interpolate_by_points(self, tth):
        l_ttheta = self.ttheta
        l_intensity = self.intensity
        tth_b = numpy.array(l_ttheta, dtype=float)
        int_b = numpy.array(l_intensity, dtype=float)
        if len(l_ttheta) == 0:
            int_1d = numpy.zeros(tth.size, dtype=float)
        else:
            int_1d = numpy.interp(tth, tth_b, int_b)
        return int_1d

    def define_points(self, pd_meas: PdMeasL, step_ttheta: float = 10.):
        ttheta = numpy.array(pd_meas.ttheta, dtype=float)
        if pd_meas.items[0].is_attribute("intensity_plus"):
            intensity = 0.5*(
                numpy.array(pd_meas.intensity_plus, dtype=float) + 
                numpy.array(pd_meas.intensity_minus, dtype=float))
        else:
            intensity = numpy.array(pd_meas.intensity, dtype=float)

        n_points = int((ttheta.max()-ttheta.min())/step_ttheta + 2)
        ttheta_bkgr = numpy.linspace(ttheta.min(), ttheta.max(), n_points, endpoint=True)
        
        flags = numpy.abs(ttheta[na, :]-ttheta_bkgr[:, na])<step_ttheta
        
        int_bkrg = numpy.array([numpy.nanmin(intensity[flag], axis=0) for flag in flags], dtype=float)
        int_bkrg = numpy.round(int_bkrg, decimals=5)
        self.numpy_ttheta = ttheta_bkgr
        self.numpy_intensity = int_bkrg
        self.items = []
        self.numpy_to_items()
        return 


"""
:Constants:


:Functions:



"""
import os
import numpy
import warnings

import pycifstar
from typing import List, Tuple

F_BSCAT = os.path.join(os.path.dirname(__file__), "bscat.tab")
BSCAT = pycifstar.to_loop(F_BSCAT)


def apply_constraint_on_cell_by_type_cell(cell, type_cell:str,
                                          it_coordinate_system_code:str):
    length_a, length_b, length_c = cell.length_a, cell.length_b, cell.length_c
    angle_alpha, angle_beta = cell.angle_alpha, cell.angle_beta
    angle_gamma = cell.angle_gamma
    if length_a is None: 
        cell.length_a = 1.
    if length_b is None: 
        cell.length_b = 1.
    if length_c is None: 
        cell.length_c = 1.
    if angle_alpha is None: 
        cell.angle_alpha = 90.
    if angle_beta is None: 
        cell.angle_beta = 90.
    if angle_gamma is None: 
        cell.angle_gamma = 90.

    if type_cell == "aP":
        pass
    elif type_cell.startswith("m"):
        cell.angle_alpha, cell.angle_alpha_refinement, cell.angle_alpha_constraint = 90., False, True
        cell.angle_gamma, cell.angle_gamma_refinement, cell.angle_gamma_constraint = 90., False, True
    elif type_cell.startswith("o"):
        cell.angle_alpha, cell.angle_alpha_refinement, cell.angle_alpha_constraint = 90., False, True
        cell.angle_beta, cell.angle_beta_refinement, cell.angle_beta_constraint = 90., False, True
        cell.angle_gamma, cell.angle_gamma_refinement, cell.angle_gamma_constraint = 90., False, True
    elif ((type_cell.startswith("t")) | (type_cell == "hP")):
        cell.length_b, cell.length_b_sigma = cell.length_a, cell.length_a_sigma
        cell.length_b_refinement, cell.length_b_constraint = False, True
        cell.angle_alpha, cell.angle_alpha_refinement, cell.angle_alpha_constraint = 90., False, True
        cell.angle_beta, cell.angle_beta_refinement, cell.angle_beta_constraint = 90., False, True
        cell.angle_gamma, cell.angle_gamma_refinement, cell.angle_gamma_constraint = 90., False, True
    elif (type_cell == "hR"):
        if it_coordinate_system_code.lower() == "h":
            cell.length_b, cell.length_b_sigma = cell.length_a, cell.length_a_sigma
            cell.length_b_refinement, cell.length_b_constraint = False, True
            cell.angle_alpha, cell.angle_alpha_refinement, cell.angle_alpha_constraint = 90., False, True
            cell.angle_beta, cell.angle_beta_refinement, cell.angle_beta_constraint = 90., False, True
            cell.angle_gamma, cell.angle_gamma_refinement, cell.angle_gamma_constraint = 120., False, True
        else:
            cell.length_b, cell.length_b_sigma = cell.length_a, cell.length_a_sigma
            cell.length_b.refinement, cell.length_b_constraint = False, True
            cell.length_c, cell.length_c_sigma = cell.length_a, cell.length_a_sigma
            cell.length_c_refinement, cell.length_c_constraint = False, True
            cell.angle_beta, cell.angle_beta_sigma = cell.angle_alpha, cell.angle_alpha_sigma
            cell.angle_beta_refinement, cell.angle_beta_constraint = False, True
            cell.angle_gamma, cell.angle_gamma_sigma = cell.angle_alph, cell.angle_alpha_sigma
            cell.angle_gamma_refinement, cell.angle_gamma_constraint = False, True
    elif type_cell.startswith("c"):
        cell.length_b, cell.length_b_sigma = cell.length_a, cell.length_a_sigma
        cell.length_b.refinement, cell.length_b_constraint = False, True
        cell.length_c, cell.length_c_sigma = cell.length_a, cell.length_a_sigma
        cell.length_c_refinement, cell.length_c_constraint = False, True
        cell.angle_alpha, cell.angle_alpha_refinement, cell.angle_alpha_constraint = 90., False, True
        cell.angle_beta, cell.angle_beta_refinement, cell.angle_beta_constraint = 90., False, True
        cell.angle_gamma, cell.angle_gamma_refinement, cell.angle_gamma_constraint = 90., False, True


def get_scat_length_neutron(type_n):
    """
    Take scat_length_neutron.
    """
    str_1 = type_n.strip().lower()
    flag_label = False
    l_hh = []
    for hh in str_1:
        if hh.isdigit() and flag_label:
            break
        elif hh.isdigit():
            l_hh.append(hh)
        else:
            l_hh.append(hh)
            flag_label = True
    str_1 = "".join(l_hh)

    flag = False
    for _1, _2 in zip(BSCAT["_atom_type_symbol"], BSCAT["_atom_type_cohb"]):
        if (_1.lower() == str_1):
            res = 0.1 * complex(_2)  # in 10**-12cm
            flag = True
        elif flag:
            break
    if not(flag):
        res = 0.
        warnings.warn(
            f"Can not find b_scat for '{type_n:}'.\n It is putted as 0.",
            UserWarning, stacklevel=2)
    return res



""" TOF

Expressions are taken from paper

J. Appl. Cryst. (1982). 15, 581-589


Functions
---------
    - 
    
"""
import numpy
import scipy
import scipy.special

inv_8ln2 = 0.18033688011112042591999058512524

na = numpy.newaxis
square = numpy.square
exp = numpy.exp
imag = numpy.imag
erfc = scipy.special.erfc
exp1 = scipy.special.exp1  # E1 = \int_{z}^{\inf} exp(-t)/t dt


def calc_hpv_eta(h_g, h_l):
    """pseudo-Voight function
    
    calculate h_pV and eta based on Gauss and Lorentz Size
    """
    h_g_2, h_l_2 = h_g**2, h_l**2
    h_g_3, h_l_3 = h_g_2*h_g, h_l_2*h_l
    h_g_4, h_l_4 = h_g_3*h_g, h_l_3*h_l
    h_g_5, h_l_5 = h_g_4*h_g, h_l_4*h_l
    c_2, c_3, c_4, c_5 = 2.69269, 2.42843, 4.47163, 0.07842
    h_pv = (h_g_5 + c_2*h_g_4*h_l + c_3*h_g_3*h_l_2 + 
            c_4*h_g_2*h_l_3 + c_5*h_g*h_l_4 + h_l_5)**0.2
    hh = h_l*1./h_pv
    eta = 1.36603*hh - 0.47719*hh**2 + 0.11116*hh**3
    return h_pv, eta


def calc_alpha(alpha0, alpha1, d):
    """alpha0+alpha1/d
    """
    return alpha0+alpha1/d


def calc_beta(beta0, beta1, d):
    """beta0+beta1/d**4
    """
    return beta0+beta1/d**4


def calc_sigma(sigma0, sigma1, d):
    """sigma0+sigma1*d
    """
    return sigma0+sigma1*d


def calc_sigma_gamma(
        d, sigma0, sigma1, sigma2, gamma0, gamma1, gamma2,
        size_g: float = 0., size_l: float = 0., strain_g: float = 0.,
        strain_l: float = 0.):
    """Calculate H_G (sigma) and H_L (gamma)
    
        H_G**2 = (sigma2+size_g)*d**4 + (sigma1+strain_g)*d**2 + sigma0
        H_L = (gamma2+size_l)*d**2 + (sigma1+strain_l)*d + sigma0
    """
    d_sq = numpy.square(d)
    d_sq_sq = numpy.square(d_sq)

    h_g_sq = (sigma2+size_g) * d_sq_sq + (sigma1+strain_l) * d_sq + sigma0
    h_l = (gamma2+size_l) * d_sq + (gamma1+strain_l) * d + gamma0
    h_g = numpy.sqrt(h_g_sq)
    return h_g, h_l


def calc_y_z_u_v(alpha, beta, sigma, delta_2d):
    """Calculate y, z, u, v

    y = (alpha * sigma**2 + delta)/(sigma * 2**0.5)
    z = (beta * sigma**2 - delta)/(sigma * 2**0.5)
    u = 0.5 * alpha * (alpha*sigma**2 + 2 delta)
    v = 0.5 * beta * (beta*sigma**2 - 2 delta)

    """
    sigma_sq = square(sigma)
    y = (alpha * sigma/(2.**0.5))[:, na] + delta_2d/(sigma*2.**0.5)[:, na]
    z = (beta * sigma/(2.**0.5))[:, na] - delta_2d/(sigma*2.**0.5)[:, na]
    u = (0.5*square(alpha)*sigma_sq)[:, na] + delta_2d*alpha[:, na]
    v = (0.5*square(beta)*sigma_sq)[:, na] - delta_2d*beta[:, na]
    return y, z, u, v

def tof_Carpenter():
    pass


def tof_Jorgensen(alpha, beta, sigma, time, time_hkl):
    norm = 0.5*alpha*beta/(alpha+beta)
    time_2d, time_hkl_2d = numpy.meshgrid(time, time_hkl, indexing="ij")
    delta_2d = time_2d-time_hkl_2d
    y, z, u, v = calc_y_z_u_v(alpha, beta, sigma, delta_2d)

    exp_u = exp(u)
    exp_v = exp(v)
    exp_u[numpy.isinf(exp_u)] = 1e200
    exp_v[numpy.isinf(exp_v)] = 1e200

    res_2d = norm[:, na] * (exp_u * erfc(y) + exp_v * erfc(z))
    return res_2d


def tof_Jorgensen_VonDreele(alpha, beta, sigma, gamma, time, time_hkl):
    two_over_pi = 2.*numpy.pi
    norm = 0.5*alpha*beta/(alpha+beta)
    time_2d, time_hkl_2d = numpy.meshgrid(time, time_hkl, indexing="ij")
    delta_2d = time_2d-time_hkl_2d

    # FIXME: it has to be checked
    # sigma = gamma*(inv_8ln2)**0.5
    h_pv, eta = calc_hpv_eta(sigma, gamma)

    y, z, u, v = calc_y_z_u_v(alpha, beta, sigma, delta_2d) 

    exp_u = exp(u)
    exp_v = exp(v)
    exp_u[numpy.isinf(exp_u)] = 1e200
    exp_v[numpy.isinf(exp_v)] = 1e200

    profile_g_2d = norm[:, na] * (exp_u * erfc(y) + exp_v * erfc(z))

    z1_2d = alpha[:, na]*delta_2d + (1j*0.5*alpha*gamma)[:, na]
    z2_2d = -beta[:, na]*delta_2d + (1j*0.5*beta*gamma)[:, na]

    fz1_2d = exp1(z1_2d)
    fz2_2d = exp1(z2_2d)

    # FIXME: check it
    fz1_2d[numpy.isnan(fz1_2d)] = 0.
    fz1_2d[numpy.isinf(fz1_2d)] = 0.
    fz2_2d[numpy.isnan(fz2_2d)] = 0.
    fz2_2d[numpy.isinf(fz2_2d)] = 0.

    oml_a_2d = -imag(fz1_2d) * two_over_pi
    oml_b_2d = -imag(fz2_2d) * two_over_pi
    profile_l_2d = norm[:, na] * (oml_a_2d + oml_b_2d)
    one_e = 1. - eta
    tof_peak_2d = one_e[:, na] * profile_g_2d + eta[:, na] * profile_l_2d
    return tof_peak_2d





"""Description of Diffrn class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"

from warnings import warn
from typing import NoReturn, List, Union
import numpy
import scipy
import scipy.misc

from cryspy.A_functions_base.matrix_operations import calc_vector_product_v1_v2_v1

from cryspy.A_functions_base.extinction import calc_extinction_sphere
from cryspy.A_functions_base.flip_ratio import calc_flip_ratio_by_structure_factors
from cryspy.A_functions_base.unit_cell import calc_sthovl_by_unit_cell_parameters, calc_volume_uc_by_unit_cell_parameters, \
    calc_eq_ccs_by_unit_cell_parameters
from cryspy.A_functions_base.structure_factor import calc_f_m_perp_by_sft

from cryspy.A_functions_base.function_1_roots import calc_roots

from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.preocedures import take_items_by_class

from cryspy.C_item_loop_classes.cl_1_chi2 import Chi2
from cryspy.C_item_loop_classes.cl_1_setup import Setup
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import \
    DiffrnRadiation
from cryspy.C_item_loop_classes.cl_2_diffrn_orient_matrix import \
    DiffrnOrientMatrix
from cryspy.C_item_loop_classes.cl_1_diffrn_refln import DiffrnReflnL
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction
from cryspy.C_item_loop_classes.cl_1_phase import Phase
from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs

from cryspy.E_data_classes.cl_1_crystal import Crystal
from cryspy.E_data_classes.cl_1_mag_crystal import MagCrystal


class Diffrn(DataN):
    """
    Single diffraction experiment with polarized neutrons.

    Data items in the DIFFRN category record details about
    single diffraction measurements.

    Methods
    -------
        - calc_iint_u_d_flip_ratio
        - calc_fr
        - calc_fm_perp_loc
        - calc_chi_sq
        - params_to_cif
        - data_to_cif
        - calc_to_cif
        - estimate_FM


    Attributes
    ----------
        - setup (mandatory)
        - diffrn_radiation (mandatory)
        - diffrn_orient_matrix (mandatory)
        - diffrn_refln (mandatory)
        - extinction
        - phase
        - refln_#phase_name
        - refln_susceptibility_#phase_name
        - refine_ls
    """

    CLASSES_MANDATORY = (Setup, DiffrnRadiation, DiffrnOrientMatrix,
                         DiffrnReflnL)
    CLASSES_OPTIONAL = (Extinction, Phase, ReflnL, ReflnSusceptibilityL,
                        RefineLs, Chi2)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "diffrn"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, data_name=None, **kwargs) -> NoReturn:
        super(Diffrn, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["data_name"] = data_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_iint_u_d_flip_ratio(
            self, index_hkl, l_crystal: List[Crystal], # or list of MagCrystal
            dict_in_out: dict = None,
            flag_internal: bool = True):
        """
        Calculate intensity for the given diffraction angle.

        Keyword Arguments
        -----------------
            h, k, l: 1D numpy array of Miller indexes
            l_crystal: list of crystal structures

        Output
        ------
            iint_u: intensity up (without scale factor)
            iint_d: intensity down (without scale factor)
            flip_ratio: flip ratio (intensity_plus/intensity_minus)

        """
        if dict_in_out is None:
            flag_dict = False
            dict_in_out_keys = []
        else:
            flag_dict = True
            dict_in_out_keys = dict_in_out.keys()

        l_label = [_.data_name for _ in l_crystal]
        if self.phase.label in l_label:
            ind = l_label.index(self.phase.label)
            crystal = l_crystal[ind]
        elif self.data_name in l_label:
            self.phase = Phase(label=self.data_name)
            ind = l_label.index(self.data_name)
            crystal = l_crystal[ind]
        else:
            warn("Crystal    not found. The first one is taken.")
            crystal = l_crystal[0]
            self.phase.label = crystal.label

        cell = crystal.cell
        unit_cell_parameters = cell.get_unit_cell_parameters()
        volume_unit_cell = cell.volume

        sthovl, dder_sthovl = calc_sthovl_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=False)
        eq_ccs, dder_eq_ccs  = calc_eq_ccs_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=False)

        setup = self.setup
        wavelength = setup.wavelength
        field_norm = setup.field
        try:
            ratio_lambdaover2 = setup.ratio_lambdaover2
            flag_lambdaover2 = True
        except AttributeError:
            ratio_lambdaover2 = None
            flag_lambdaover2 = False

        diffrn_orient_matrix = self.diffrn_orient_matrix
        u_matrix = diffrn_orient_matrix.u
        magnetic_field = field_norm*u_matrix[2, :]
        matrix_u = numpy.array([
            u_matrix[0,0], u_matrix[0,1], u_matrix[0,2],
            u_matrix[1,0], u_matrix[1,1], u_matrix[1,2],
            u_matrix[2,0], u_matrix[2,1], u_matrix[2,2]], dtype=float)
        diffrn_radiation = self.diffrn_radiation
        polarization = diffrn_radiation.polarization
        flipper_efficiency = diffrn_radiation.efficiency

        if self.is_attribute("extinction"):
            extinction = self.extinction
            model_extinction = extinction.model
            radius = extinction.radius
            mosaicity = extinction.mosaicity
        else:
            model_extinction = ""
            radius = None
            mosaicity = None
        if flag_lambdaover2:
            index_2hkl = 2 * index_hkl

        if flag_internal:
            refln = crystal.calc_refln(index_hkl)
            f_nucl = numpy.array(refln.f_calc, dtype=complex)
        else:
            f_nucl = crystal.calc_f_nucl(index_hkl)

        if flag_dict:
            dict_in_out["phase_name"] = crystal.get_name()
            dict_in_out["index_hkl"] = index_hkl
            dict_in_out["f_nucl"] = f_nucl
        
        if flag_lambdaover2:
            f_nucl_2hkl = crystal.calc_f_nucl(index_2hkl)
            if flag_dict:
                dict_in_out["f_nucl_2hkl"] = f_nucl_2hkl
        else:
            f_nucl_2hkl = None
        
        refln_s = None
        if isinstance(crystal, Crystal):
            if flag_internal:
                refln_s = crystal.calc_refln_susceptibility(index_hkl)
                sft_ccs = numpy.stack([
                    refln_s.numpy_chi_11_calc, refln_s.numpy_chi_12_calc, refln_s.numpy_chi_13_calc,
                    refln_s.numpy_chi_21_calc, refln_s.numpy_chi_22_calc, refln_s.numpy_chi_23_calc,
                    refln_s.numpy_chi_31_calc, refln_s.numpy_chi_32_calc, refln_s.numpy_chi_33_calc],
                    axis=0)
                fm_perp_loc, dder_fm_perp_loc = calc_f_m_perp_by_sft(
                    sft_ccs, magnetic_field, eq_ccs, flag_magnetic_field=False, flag_eq_ccs=False, flag_sft_ccs=False) 
            else:
                fm_perp_loc = crystal.calc_f_m_perp(index_hkl, magnetic_field, dict_in_out=dict_in_out)
                if flag_dict:
                    sft_ccs = dict_in_out["sft_ccs"]
            if flag_dict:
                dict_in_out["f_m_perp"] = fm_perp_loc
        elif isinstance(crystal, MagCrystal):
            fm_perp_loc = crystal.calc_f_mag_perp(index_hkl)

        if flag_lambdaover2:
            if isinstance(crystal, Crystal):
                dict_in_out_2 = {"phase_name": crystal.get_name()}
                fm_perp_loc_2hkl = crystal.calc_f_m_perp(index_2hkl, magnetic_field, dict_in_out=dict_in_out_2)
                if flag_dict:
                    dict_in_out["sft_ccs_2hkl"] = dict_in_out_2.pop("sft_ccs")
            elif isinstance(crystal, MagCrystal):
                fm_perp_loc_2hkl = crystal.calc_f_mag_perp(index_2hkl)
                if flag_dict:
                    dict_in_out["f_m_perp_2hkl"] = fm_perp_loc_2hkl
        else:
            fm_perp_loc_2hkl = None


        def func_extinction(f_sq, flag_f_sq: bool = False):
            cos_2theta = 1.-2*numpy.square(sthovl * wavelength)
            flag_radius = False
            flag_mosaicity = False
            flag_volume_unit_cell = False
            flag_cos_2theta = False
            flag_wavelength = False
            return calc_extinction_sphere(
                f_sq, radius, mosaicity, volume_unit_cell, cos_2theta, wavelength,
                model_extinction, flag_f_sq=flag_f_sq, flag_radius=flag_radius,
                flag_mosaicity=flag_mosaicity,
                flag_volume_unit_cell=flag_volume_unit_cell,
                flag_cos_2theta=flag_cos_2theta,
                flag_wavelength=flag_wavelength)

        flag_polarization = False
        flag_flipper = False
        flag_f_n = False
        flag_f_m_perp = False
        flag_c_lambda2 = False
        flag_f_n_2h = False 
        flag_f_m_perp_2h = False
        flip_ratio, dder_fr = calc_flip_ratio_by_structure_factors(
            polarization, flipper_efficiency, f_nucl, fm_perp_loc, matrix_u,
            func_extinction=func_extinction,
            c_lambda2=ratio_lambdaover2, f_nucl_2hkl=f_nucl_2hkl, f_m_perp_2hkl=fm_perp_loc_2hkl,
            flag_beam_polarization=flag_polarization, flag_flipper_efficiency=flag_flipper,
            flag_f_nucl=flag_f_n, flag_f_m_perp=flag_f_m_perp,
            flag_c_lambda2=flag_c_lambda2,
            flag_f_nucl_2hkl=flag_f_n_2h, flag_f_m_perp_2hkl=flag_f_m_perp_2h,
            dict_in_out=dict_in_out, flag_use_precalculated_data=flag_internal)


        dder = {}

        if flag_internal:
            refln.loop_name = crystal.data_name
            self.add_items([refln, ])
            if refln_s is not None:
                refln_s.loop_name = crystal.data_name
                self.add_items([refln_s, ])


        return flip_ratio, dder


    def calc_fr(self, cell, f_nucl, f_m_perp, delta_f_nucl=None,
                delta_f_m_perp=None):
        """
        Calculate flip ratio.

        By given nuclear structure factor and
        perpendicular component of magnetic structure factor
        (in 10**-12 cm)
        cell (only for extinction correction)
        f_nucl [hkl]
        f_m_perp [3, hkl]
        delta_f_nucl [hkl, parameters]
        delta_f_m_perp [hkl, parameters]
        """


        diffrn_radiation = self.diffrn_radiation
        polarization = diffrn_radiation.polarization
        flipper = diffrn_radiation.efficiency
        m_u = self.diffrn_orient_matrix.u

        unit_cell_parameters = cell.get_unit_cell_parameters()
        index_hkl = numpy.array([
            self.diffrn_refln.index_h,
            self.diffrn_refln.index_k,
            self.diffrn_refln.index_l], dtype=int)
        sthovl = calc_sthovl_by_unit_cell_parameters(index_hkl, unit_cell_parameters)[0]

        flag_polarization = False
        flag_flipper = False
        flag_f_n = False
        flag_f_m_perp = True

        c_lambda2 = None
        f_n_2h = None
        f_m_perp_2h = None
        flag_c_lambda2 = False
        flag_f_n_2h = False
        flag_f_m_perp_2h = False

        def func_extinction(f_sq, flag_f_sq: bool = False):
            model = self.extinction.model
            radius = self.extinction.radius
            mosaicity = self.extinction.mosaicity
            volume_unit_cell = calc_volume_uc_by_unit_cell_parameters(unit_cell_parameters)[0]
            wavelength = self.setup.wavelength
            cos_2theta = 1.-2*numpy.square(sthovl * wavelength)
            flag_radius = True
            flag_mosaicity = True
            flag_volume_unit_cell = True
            flag_cos_2theta = True
            flag_wavelength = True
            return calc_extinction_sphere(
                f_sq, radius, mosaicity, volume_unit_cell, cos_2theta, wavelength,
                model, flag_f_sq=flag_f_sq, flag_radius=flag_radius,
                flag_mosaicity=flag_mosaicity,
                flag_volume_unit_cell=flag_volume_unit_cell,
                flag_cos_2theta=flag_cos_2theta,
                flag_wavelength=flag_wavelength)

        f_r_m, dder = calc_flip_ratio_by_structure_factors(
            polarization, flipper, f_nucl, f_m_perp, m_u, func_extinction=func_extinction,
            c_lambda2=c_lambda2, f_n_2h=f_n_2h, f_m_perp_2h=f_m_perp_2h,
            flag_polarization=flag_polarization, flag_flipper=flag_flipper,
            flag_f_n=flag_f_n, flag_f_m_perp=flag_f_m_perp,
            flag_c_lambda2=flag_c_lambda2,
            flag_f_n_2h=flag_f_n_2h, flag_f_m_perp_2h=flag_f_m_perp_2h)

        delta_f_r_m = (
            numpy.expand_dims(dder["f_m_perp_real"], 2)*
            delta_f_m_perp.real +
            numpy.expand_dims(dder["f_m_perp_imag"], 2)*
            delta_f_m_perp.imag).sum(axis=0)[:3]

        return f_r_m, delta_f_r_m


    def calc_chi_sq(
            self, l_crystal: List[Crystal], # or MagCrystal
            flag_internal=True, dict_in_out: dict = None):
        """
        Calculate chi square.

        Keyword Arguments
        -----------------
            - l_crystal: a list of Crystal objects of cryspy library
            - flag_internal: a flag to calculate internal objects
              (default is True)

        Output arguments
        ----------------
            - chi_sq_val: chi square of flip ratio
              (Sum_i ((y_e_i - y_m_i) / sigma_i)**2)
            - n: number of measured reflections
        """
        diffrn_refln = self.diffrn_refln
        index_h = diffrn_refln.numpy_index_h
        index_k = diffrn_refln.numpy_index_k
        index_l = diffrn_refln.numpy_index_l
        index_hkl = numpy.stack([index_h, index_k, index_l], axis=0)
        fr_exp = diffrn_refln.numpy_fr
        fr_sigma = diffrn_refln.numpy_fr_sigma
        fr_mod, dder = self.calc_iint_u_d_flip_ratio(index_hkl, l_crystal, flag_internal=flag_internal, dict_in_out=dict_in_out)

        if flag_internal:
            diffrn_refln.numpy_fr_calc = fr_mod
            # diffrn_refln.numpy_intensity_plus_calc = int_u_mod
            # diffrn_refln.numpy_intensity_minus_calc = int_d_mod
            diffrn_refln.numpy_to_items()

        flag_in = numpy.logical_not(diffrn_refln.numpy_excluded)

        chi_sq = ((fr_mod[flag_in]-fr_exp[flag_in])/fr_sigma[flag_in])**2
        chi_sq_val = (chi_sq[numpy.logical_not(numpy.isnan(chi_sq))]).sum()
        n = numpy.logical_not(numpy.isnan(chi_sq)).sum()

        if flag_internal:
            refine_ls = RefineLs(number_reflns=n,
                                 goodness_of_fit_all=chi_sq_val/float(n),
                                 weighting_scheme="sigma")
            self.refine_ls = refine_ls
        return chi_sq_val, n

    def params_to_cif(self, separator="_") -> str:
        """Save parameters in cif format."""
        ls_out = []
        l_cls = (Setup, DiffrnRadiation, DiffrnOrientMatrix, Extinction, Phase)
        l_obj = [item for item in self.items if type(item) in l_cls]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_obj])
        return "\n".join(ls_out)

    def data_to_cif(self, separator="_") -> str:
        """Save data in cif format."""
        ls_out = []
        l_cls = (DiffrnReflnL, )
        l_obj = [item for item in self.items if type(item) in l_cls]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_obj])
        return "\n".join(ls_out)

    def calc_to_cif(self, separator="_", flag=False) -> str:
        """Save calculations in cif format."""
        ls_out = []
        l_cls = (RefineLs, ReflnL, ReflnSusceptibilityL)
        l_obj = [item for item in self.items if type(item) in l_cls]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_obj])
        return "\n".join(ls_out)

    def estimate_FM(self, crystal, max_abs_f_mag: float = 5,
                    delta_f_mag: float = 0.001, flag_lambdaover2_correction=False):
        """
        Estimate magnetic structure factor.

        The method calculates magnetic structure factors (only real part) from
        flipping ratios based on give crystal structure.
        The crystal structure should be centrosymmetrical
        """
        setup = self.setup
        wavelength = setup.wavelength
        field_norm = setup.field

        diffrn_refln = self.diffrn_refln
        index_h = diffrn_refln.numpy_index_h
        index_k = diffrn_refln.numpy_index_k
        index_l = diffrn_refln.numpy_index_l
        index_hkl = numpy.stack([index_h, index_k, index_l], axis=0)
        fr_exp = diffrn_refln.numpy_fr
        fr_sigma = diffrn_refln.numpy_fr_sigma

        diffrn_radiation = self.diffrn_radiation
        polarization = diffrn_radiation.polarization
        flipper_efficiency = diffrn_radiation.efficiency

        if self.is_attribute("extinction"):
            extinction = self.extinction
            model_extinction = extinction.model
            radius = extinction.radius
            mosaicity = extinction.mosaicity
        else:
            model_extinction = ""
            radius = None
            mosaicity = None

        diffrn_orient_matrix = self.diffrn_orient_matrix
        u_matrix = diffrn_orient_matrix.u
        magnetic_field = field_norm * u_matrix[2, :] 

        refln = crystal.calc_refln(index_hkl)
        f_nucl = numpy.array(refln.f_calc, dtype=complex)

        cell = crystal.cell
        unit_cell_parameters = cell.get_unit_cell_parameters()
        sthovl = cell.calc_sthovl(index_hkl[0], index_hkl[1], index_hkl[2])
        eq_ccs, dder_eq_ccs = calc_eq_ccs_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=False)
        

        if (setup.is_attribute("ratio_lambdaover2") &
                flag_lambdaover2_correction):
            ratio_lambdaover2 = setup.ratio_lambdaover2
            flag_lambdaover2 = True
        else:
            ratio_lambdaover2 = None
            flag_lambdaover2 = False

        if (flag_lambdaover2 & flag_lambdaover2_correction):
            index_2hkl = 2 * index_hkl
            f_nucl_2hkl = crystal.calc_f_nucl(index_2hkl)
        else:
            f_nucl_2hkl = None

        if (flag_lambdaover2 & flag_lambdaover2_correction):
            sft_2hkl, dder_sft_2hkl = crystal.calc_structure_factor_tensor_ccs(index_2hkl)

            fm_perp_loc_2hkl, dder_f_m_perp = calc_f_m_perp_by_sft(
                sft_2hkl, magnetic_field, eq_ccs,
                flag_sft_ccs=False,
                flag_magnetic_field=False,
                flag_eq_ccs=False)

        else:
            fm_perp_loc_2hkl = None

        def func_extinction(f_sq, flag_f_sq: bool = False):
            model = self.extinction.model
            radius = self.extinction.radius
            mosaicity = self.extinction.mosaicity
            volume_unit_cell = calc_volume_uc_by_unit_cell_parameters(unit_cell_parameters)[0]
            wavelength = self.setup.wavelength
            cos_2theta = 1.-2*numpy.square(sthovl * wavelength)
            flag_radius = True
            flag_mosaicity = True
            flag_volume_unit_cell = True
            flag_cos_2theta = True
            flag_wavelength = True
            return calc_extinction_sphere(
                f_sq, radius, mosaicity, volume_unit_cell, cos_2theta, wavelength,
                model, flag_f_sq=flag_f_sq, flag_radius=flag_radius,
                flag_mosaicity=flag_mosaicity,
                flag_volume_unit_cell=flag_volume_unit_cell,
                flag_cos_2theta=flag_cos_2theta,
                flag_wavelength=flag_wavelength)

        def function_to_find_fm(f_mag):
            """Calc flip ratio.

            Parameters
            ----------
            f_mag : float
                in muB.

            Returns
            -------
            flip_ratio : float
                flip ratio.
            """
            mag_loc_1 = u_matrix[2, 0]*f_mag*0.2695
            mag_loc_2 = u_matrix[2, 1]*f_mag*0.2695
            mag_loc_3 = u_matrix[2, 2]*f_mag*0.2695
            mag_loc = numpy.stack([mag_loc_1, mag_loc_2, mag_loc_3], axis=0)
            f_m_perp, dder_fm_perp_loc = calc_vector_product_v1_v2_v1(eq_ccs, mag_loc, flag_v1=False, flag_v2=False)


            f_r, dder_f_r = calc_flip_ratio_by_structure_factors(
                       polarization, flipper_efficiency, f_nucl, f_m_perp, u_matrix, func_extinction=func_extinction,
                       c_lambda2=ratio_lambdaover2, f_n_2h=f_nucl_2hkl, f_m_perp_2h=fm_perp_loc_2hkl,
                       flag_polarization=False, flag_flipper=False,
                       flag_f_n=False, flag_f_m_perp=False,
                       flag_c_lambda2=False,
                       flag_f_n_2h=False, flag_f_m_perp_2h=False)

            return f_r

        lFMans, lFMansMin = [], []
        lFMansMin_sigma = []
        for _ind in range(fr_exp.size):
            def f(x):
                return (function_to_find_fm(x)[_ind] - fr_exp[_ind])
            FMans = calc_roots(f, -1*max_abs_f_mag-delta_f_mag, max_abs_f_mag+delta_f_mag)
            lFMans.append(FMans)
            if len(FMans) == 0:
                lFMansMin.append(None)
                lFMansMin_sigma.append(None)
            else:
                FMans_min = min(FMans)
                lFMansMin.append(FMans_min)
                der1 = scipy.misc.derivative(f, FMans_min, 0.1, 1)
                f_mag_sigma = abs(fr_sigma[_ind] / der1)
                lFMansMin_sigma.append(f_mag_sigma)

        ls_out = [f"loop_{crystal.data_name:}"]
        ls_out.append("_estimate_index_h\n_estimate_index_k")
        ls_out.append("_estimate_index_l\n_estimate_F_M\n_estimate_F_M_sigma")
        for ind_h, ind_k, ind_l, f_mag, f_sig in \
                zip(index_h, index_k, index_l, lFMansMin, lFMansMin_sigma):
            ls_out.append(f"{ind_h:} {ind_k:} {ind_l:} {f_mag:}  {f_sig:}")

        res = LoopN.from_cif("\n".join(ls_out))

        return res

    def apply_constraints(self):
        """Apply constraints."""
        pass

    def plots(self):
        l_res = []
        if self.is_attribute("diffrn_refln"):
            diffrn_refln = self.diffrn_refln
            flag_asymmetry = False
            if self.is_attribute("chi2"):
                chi2 = self.chi2
                if chi2.is_attribute("asymmetry"):
                    flag_asymmetry = chi2.asymmetry

            if flag_asymmetry:
                fig_ax = diffrn_refln.plot_asymmetry_vs_asymmetry_calc()
            else:
                fig_ax = diffrn_refln.plot_fr_vs_fr_calc()

            if fig_ax is not None:
                fig, ax = fig_ax
                ax.set_title(self.data_name + " - "+ax.title.get_text())
                l_res.append((fig, ax))
        return l_res

    def report(self):
        if self.is_attribute("diffrn_refln"):
            diffrn_refln = self.diffrn_refln
            rep_1 = diffrn_refln.report_agreement_factor_exp()
            rep_2 = ""
            if self.is_attribute("diffrn_orient_matrix"):
                ub_matrix = self.diffrn_orient_matrix
                if ub_matrix.is_defined():
                    cell = ub_matrix.cell
                    rep_2 = diffrn_refln.report_chi_sq_exp(cell=cell)
            if rep_2 == "":
                rep_2 = diffrn_refln.report_chi_sq_exp()
            return rep_1 + "\n" + rep_2
        return  ""

    def get_dictionary(self):
        """Form dictionary. See documentation moduel CrysPy using Jupyter notebook.
        """
        
        self.form_object()
        ddict = {}
        setup, diffrn_refln, diffrn_orient_matrix = None, None, None
        extinction, diffrn_radiation = None, None
        phase, chi2 = None, None

        l_obj = take_items_by_class(self, (Setup, ))
        if len(l_obj) > 0:
            setup = l_obj[0]

        l_obj = take_items_by_class(self, (DiffrnReflnL, ))
        if len(l_obj) > 0:
            diffrn_refln = l_obj[0]

        l_obj = take_items_by_class(self, (DiffrnOrientMatrix, ))
        if len(l_obj) > 0:
            diffrn_orient_matrix = l_obj[0]

        l_obj = take_items_by_class(self, (Extinction, ))
        if len(l_obj) > 0:
            extinction = l_obj[0]

        l_obj = take_items_by_class(self, (DiffrnRadiation, ))
        if len(l_obj) > 0:
            diffrn_radiation = l_obj[0]

        l_obj = take_items_by_class(self, (Phase, ))
        if len(l_obj) > 0:
            phase = l_obj[0]

        l_obj = take_items_by_class(self, (Chi2, ))
        if len(l_obj) > 0:
            chi2 = l_obj[0]

        ddict["name"] = self.data_name
        ddict["type_name"] = self.get_name()
        if setup is not None:
            field = setup.field
            wavelength = setup.wavelength
            ddict["wavelength"] = numpy.array([wavelength], dtype=float)
            ddict["flags_wavelength"] = numpy.array([setup.wavelength_refinement], dtype=bool)
            if setup.is_attribute("ratio_labmdaover2"):
                ddict["c_lambda2"] = numpy.array([setup.ratio_lambdaover2], dtype=float)
                ddict["flags_c_lambda2"] = numpy.array([setup.ratio_lambdaover2_refinement], dtype=float)

        if diffrn_refln is not None:
            index_hkl = numpy.array([
                diffrn_refln.index_h, diffrn_refln.index_k,
                diffrn_refln.index_l], dtype=float)
            refln_fr_es = numpy.array([
                diffrn_refln.fr, diffrn_refln.fr_sigma], dtype=float)
            refln_fr_excl = numpy.array(diffrn_refln.excluded, dtype=bool)

            ddict["index_hkl"] = index_hkl
            ddict["flip_ratio_es"] = refln_fr_es
            ddict["flip_ratio_excluded"] = refln_fr_excl

        if diffrn_orient_matrix is not None:
            u_matrix = diffrn_orient_matrix.u
            e_up = numpy.array([u_matrix[2,0], u_matrix[2,1], u_matrix[2,2]], dtype=float)
            
            ddict["matrix_u"] = numpy.array([
                diffrn_orient_matrix.u_11, diffrn_orient_matrix.u_12,
                diffrn_orient_matrix.u_13, diffrn_orient_matrix.u_21,
                diffrn_orient_matrix.u_22, diffrn_orient_matrix.u_23,
                diffrn_orient_matrix.u_31, diffrn_orient_matrix.u_32,
                diffrn_orient_matrix.u_33], dtype = float)
            ddict["matrix_ub"] = numpy.array([
                diffrn_orient_matrix.ub_11, diffrn_orient_matrix.ub_12,
                diffrn_orient_matrix.ub_13, diffrn_orient_matrix.ub_21,
                diffrn_orient_matrix.ub_22, diffrn_orient_matrix.ub_23,
                diffrn_orient_matrix.ub_31, diffrn_orient_matrix.ub_32,
                diffrn_orient_matrix.ub_33], dtype = float)

        if extinction is not None:
            model_extinction = extinction.model
            radius = extinction.radius
            mosaicity = extinction.mosaicity
            ddict["extinction_model"] = model_extinction
            ddict["extinction_radius"] = numpy.array([radius], dtype=float)
            ddict["extinction_mosaicity"] = numpy.array([mosaicity], dtype=float)
            ddict["flags_extinction_radius"] = numpy.array([extinction.radius_refinement], dtype=bool)
            ddict["flags_extinction_mosaicity"] = numpy.array([extinction.mosaicity_refinement], dtype=bool)

        if diffrn_radiation is not None:
            beam_polarization = diffrn_radiation.polarization
            flipper_efficiency = diffrn_radiation.efficiency
            ddict["beam_polarization"] = numpy.array([beam_polarization], dtype=float)
            ddict["flipper_efficiency"] = numpy.array([flipper_efficiency], dtype=float)
            ddict["flags_beam_polarization"] = numpy.array([diffrn_radiation.polarization_refinement], dtype=bool)
            ddict["flags_flipper_efficiency"] = numpy.array([diffrn_radiation.efficiency_refinement], dtype=bool)

        if phase is not None:
            ddict["phase_label"] = numpy.array([phase.label], dtype=str)
        else:
            ddict["phase_label"] = numpy.array([""], dtype=str)

        if chi2 is not None:
            if chi2.is_attribute("asymmetry"):
                ddict["flag_asymmetry"] = numpy.array([chi2.asymmetry], dtype=bool)
            else:
                ddict["flag_asymmetry"] = numpy.array([False], dtype=bool)
        else:
            ddict["flag_asymmetry"] = numpy.array([False], dtype=bool)
            
        keys = ddict.keys()
        if ((setup is not None) and ("matrix_u" in keys)):
            ddict["magnetic_field"] = field*ddict["matrix_u"][6:]
        return ddict


    def take_parameters_from_dictionary(self, ddict_diffrn, l_parameter_name: list=None, l_sigma: list=None):
        keys = ddict_diffrn.keys()
        if l_parameter_name is not None:
            parameter_label = [hh[0] for hh in l_parameter_name]
        else:
            parameter_label = []
        if "beam_polarization" in keys:
            self.diffrn_radiation.polarization = ddict_diffrn["beam_polarization"]
        if "flipper_efficiency" in keys:
            self.diffrn_radiation.efficiency = ddict_diffrn["flipper_efficiency"]
        if "extinction_mosaicity" in keys:
            self.extinction.mosaicity = ddict_diffrn["extinction_mosaicity"]
        if "extinction_radius" in keys:
            self.extinction.radius = ddict_diffrn["extinction_radius"]


        if len(parameter_label) > 0:
            for name, sigma in zip(l_parameter_name, l_sigma):
                if name[0] == "wavelength":
                    self.setup.wavelength_sigma = float(sigma)
                if name[0] == "beam_polarization":
                    self.diffrn_radiation.polarization_sigma = float(sigma)
                if name[0] == "flipper_efficiency":
                    self.diffrn_radiation.efficiency_sigma = float(sigma)
                if name[0] == "extinction_mosaicity":
                    self.extinction.mosaicity_sigma = float(sigma)
                if name[0] == "extinction_radius":
                    self.extinction.radius_sigma = float(sigma)

        if (("iint_plus" in keys) and ("iint_minus" in keys)):
            diffrn_refln = self.diffrn_refln
            diffrn_refln.numpy_fr_calc = ddict_diffrn["iint_plus"]/ddict_diffrn["iint_minus"]
            diffrn_refln.numpy_to_items()




import numpy
import scipy
import scipy.special
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary
from cryspy.A_functions_base.function_1_tof import tof_Jorgensen, \
    tof_Jorgensen_VonDreele, calc_hpv_eta, calc_sigma_gamma

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class TOFProfile(ItemN):
    """Profile parameters for time-of-flight experiment.

    Attributes
    ----------
        - sigma0, sigma1, sigma2, gamma0, gamma1, gamma2 
          alpha0, alpha1, beta0, beta1 (mandatory)
        - size_g, size_l, strain_g, strain_l, peak_shape (optional)

    The variance of the Gaussian component of the peak shape in TOF neutrons is
    given by:
 
        H_G**2 = (sigma_2+size_g)*d**4 + (sigma_1+strain_g)*d**2 + sigma_0
        H_L = (gamma_2+size_l)*d**2 + (gamma_1+strain_l)*d + gamma_0
        
    where d is the d-spacing in angstroms.

    Units
    -----
        sigma2, gamma2 are (microsecs/Ang.**2)**2;
        sigma1, gamma1 are (microsecs/Ang.**2);
        sigma0, gamma0 are (microsecs**2);

    """
    ATTR_MANDATORY_NAMES = ("alpha0", "alpha1", "beta0", "beta1", "sigma0",
                            "sigma1")
    ATTR_MANDATORY_TYPES = (float, float, float, float, float, float)
    ATTR_MANDATORY_CIF = ("alpha0", "alpha1", "beta0", "beta1", "sigma0",
                          "sigma1")

    ATTR_OPTIONAL_NAMES = (
        "sigma2", "peak_shape", "size_g", "size_l", "strain_g", "strain_l",
        "gamma2", "gamma1", "gamma0")
    ATTR_OPTIONAL_TYPES = (float, str, float, float, float, float, float,
                           float, float)
    ATTR_OPTIONAL_CIF = (
        "sigma2", "peak_shape", "size_g", "size_l", "strain_g", "strain_l",
        "gamma2", "gamma1", "gamma0")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("sigma2", "sigma1", "sigma0", "gamma2", "gamma1", "gamma0",
                "alpha0", "alpha1", "beta0", "beta1", "size_g", "size_l",
                "strain_g", "strain_l")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"sigma0": "{:.5f}", "sigma1": "{:.5f}", "sigma2": "{:.5f}",
                 "gamma0": "{:.5f}", "gamma1": "{:.5f}", "gamma2": "{:.5f}",
                 "alpha0": "{:.5f}", "alpha1": "{:.5f}", "beta0": "{:.5f}",
                 "beta1": "{:.5f}", "size_g": "{:.5f}", "size_l": "{:.5f}",
                 "strain_g": "{:.5f}", "strain_l": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {"peak_shape": ["pseudo-Voigt", "Gauss"]}

    # default values for the parameters
    D_DEFAULT = {"peak_shape": "pseudo-Voigt"}

    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "tof_profile"

    def __init__(self, **kwargs) -> NoReturn:
        super(TOFProfile, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_hpveta(self, h_g, h_l):
        """
        ttheta in radians, could be array
        pseudo-Voight function
        """
        h_pv, eta = calc_hpv_eta(h_g, h_l)
        return h_pv, eta

    def calc_agbg(self, h_pv):
        a_g = (2./h_pv)*(numpy.log(2.)/numpy.pi)**0.5
        b_g = 4*numpy.log(2)/(h_pv**2)
        return a_g, b_g

    def calc_albl(self, h_pv):
        a_l = 2./(numpy.pi*h_pv)
        b_l = 4./(h_pv**2)
        return a_l, b_l


    def calc_peak_shape_function(
            self, d, time, time_hkl, size_g: float = 0., strain_g: float = 0.,
            size_l: float = 0., strain_l: float = 0.):
        """Calculate peak-shape function F(DELTA)

        For Gauss peak-shape:
            F(DELTA) = alpha * beta / (alpha+beta) * [exp(u) erfc(y) + exp(v) erfc(z)]

        For pseudo-Voigt peak-shape:
            F(DELTA) = ...

        alpha = alpha0 + alpha1 / d
        beta = beta0 + beta1 / d**4
        sigma = sigma0 + sigma1 * d

        u = alpha/2 * (alpha * sigma**2 + 2 DELTA)
        v = beta/2 * (beta * sigma**2 - 2 DELTA)
        y = (alpha * sigma**2 + DELTA)/(sigma * 2**0.5)
        z = (beta * sigma**2 - DELTA)/(sigma * 2**0.5)

        DELTA = time - time_hkl

        """
        
        alpha = self.alpha0 + self.alpha1 / d
        beta = self.beta0 + self.beta1 / d**4
        
        if self.peak_shape == "Gauss":
            sigma = self.sigma0 + self.sigma1 * d
            res_2d = tof_Jorgensen(alpha, beta, sigma, time, time_hkl)
        else:  # self.peak_shape == "pseudo-Voigt"
            if self.is_attribute("size_g"):
                size_g_s = self.size_g+size_g
            else:
                size_g_s = size_g
            if self.is_attribute("size_l"):
                size_l_s = self.size_l+size_l
            else:
                size_l_s = size_l
            if self.is_attribute("strain_g"):
                strain_g_s = self.strain_g+strain_g
            else:
                strain_g_s = strain_g
            if self.is_attribute("strain_l"):
                strain_l_s = self.strain_l+strain_l
            else:
                strain_l_s = strain_l

            sigma, gamma = calc_sigma_gamma(
                d, self.sigma0, self.sigma1, self.sigma2, self.gamma0,
                self.gamma1, self.gamma2, size_g=size_g_s, size_l=size_l_s,
                strain_g=strain_g_s, strain_l=strain_l_s)
            
            res_2d = tof_Jorgensen_VonDreele(
                alpha, beta, sigma, gamma, time, time_hkl)
        return res_2d

class TOFProfileL(LoopN):
    """Profile parameters for time-of-flight experiment.

    """
    ITEM_CLASS = TOFProfile
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(TOFProfileL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name
   



"""Description of Pd class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"

import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_2_crystallography_base import \
    calc_cos_ang

from cryspy.A_functions_base.matrix_operations import calc_m1_m2_m1t
from cryspy.A_functions_base.unit_cell import calc_matrix_t

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.preocedures import take_items_by_class

from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import \
    DiffrnRadiation
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction
from cryspy.C_item_loop_classes.cl_1_phase import PhaseL
from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs
from cryspy.C_item_loop_classes.cl_1_chi2 import Chi2
from cryspy.C_item_loop_classes.cl_1_range import Range
from cryspy.C_item_loop_classes.cl_1_setup import Setup
from cryspy.C_item_loop_classes.cl_1_texture import Texture, TextureL
from cryspy.C_item_loop_classes.cl_1_exclude import ExcludeL

from cryspy.C_item_loop_classes.cl_1_tof_background import TOFBackground
from cryspy.C_item_loop_classes.cl_1_tof_intensity_incident import \
    TOFIntensityIncident
from cryspy.C_item_loop_classes.cl_1_tof_meas import TOFMeasL
from cryspy.C_item_loop_classes.cl_1_tof_parameters import TOFParameters
from cryspy.C_item_loop_classes.cl_1_tof_peak import TOFPeakL
from cryspy.C_item_loop_classes.cl_1_tof_proc import TOFProcL
from cryspy.C_item_loop_classes.cl_1_tof_profile import TOFProfile

from cryspy.E_data_classes.cl_1_crystal import Crystal
from cryspy.E_data_classes.cl_1_mag_crystal import MagCrystal


class TOF(DataN):
    """
    Time-of-flight powder diffraction (polarized or unpolarized neutrons, 1d).

    Data items in the TOF category record details about
    powder diffraction measurements by Time of FLight.

    Methods
    -------
        - calc_profile

    Attributes
    ----------
        - setup, tof_profile, phase, tof_background, tof_meas (mandatory)
        - diffrn_radiation, chi2, range, extinction, 
          texture, exclude, tof_proc, tof_peak, refine_ls, refln_#phase_name
          refln_susceptibility_#phase_name (optional)
    """

    CLASSES_MANDATORY = (TOFParameters, TOFProfile, PhaseL, TOFBackground,
                         TOFMeasL)
    CLASSES_OPTIONAL = (DiffrnRadiation, Chi2, Range,
                        Texture, TOFIntensityIncident, ExcludeL, TOFProcL,
                        TOFPeakL, RefineLs, ReflnL, ReflnSusceptibilityL, Setup)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "tof"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, data_name=None, **kwargs) -> NoReturn:
        super(TOF, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["data_name"] = data_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_profile(self, time, l_crystal, flag_internal: bool = True,
                     flag_polarized: bool = True):
        """Calculate intensity for the given diffraction angles.

        Arguments
        ---------
            - time: 1D numpy array of time in micro seconds (???)
            - l_crystal: a list of Crystal objects of cryspy library
            - l_peak_in: precalculated data about integrated intensities
              (used to speed up the calculations)
            - l_refln_in: precalculated data about nuclear structure factors
              (used to speed up the calculations)
            - l_refln_susceptibility_in: precalculated data about
              (used to speed up the calculations)
            - flag_internal: a flag to calculate or to use internal objects.
                   It should be True if user call the function.
                   It's True by default.

        Output
        ------
            - proc: output profile
            - l_peak: data about peaks
            - l_refln: data about nuclear structure factor
        """
        if flag_internal:
            d_internal_val = {}
            self.d_internal_val = d_internal_val
        else:
            try:
                d_internal_val = self.d_internal_val
            except AttributeError:
                d_internal_val = {}
                self.d_internal_val = d_internal_val

        tof_proc = TOFProcL()
        tof_proc.numpy_time = time

        try:
            tof_background = self.tof_background
            int_bkgd = tof_background.calc_background(time)
        except AttributeError:
            int_bkgd = 0.*time

        tof_proc.numpy_intensity_bkg_calc = int_bkgd

        tof_parameters = self.tof_parameters
        # tof_profile = self.tof_profile
        
        d = tof_parameters.calc_d_by_time(time)
        d_min, d_max = tof_parameters.calc_d_min_max(time)
        sthovl_min = 0.5/d_max
        sthovl_max = 0.5/d_min

        try:
            texture = self.texture
            h_ax, k_ax, l_ax = texture.h_ax, texture.k_ax, texture.l_ax
            g_1, g_2 = texture.g_1, texture.g_2
        except AttributeError:
            texture = None

        res_u_1d = numpy.zeros(time.shape[0], dtype=float)
        res_d_1d = numpy.zeros(time.shape[0], dtype=float)

        phase = self.phase

        for phase_item in phase.items:
            phase_label = phase_item.label
            phase_scale = phase_item.scale
            try:
                phase_igsize = phase_item.igsize
                if phase_igsize is None: phase_igsize = 0. # temporary solution
            except AttributeError:
                phase_igsize = 0.

            for i_crystal, crystal in enumerate(l_crystal):
                if crystal.data_name.lower() == phase_label.lower():
                    ind_cry = i_crystal
                    break
            if ind_cry is None:
                raise AttributeError(
                    f"Crystal with name '{crystal.data_name:}' is not found.")
                return
            crystal = l_crystal[ind_cry]

            scale = phase_scale

            try:
                peak = d_internal_val[f"peak_{crystal.data_name:}"]
                index_h = peak.numpy_index_h
                index_k = peak.numpy_index_k
                index_l = peak.numpy_index_l
                mult = peak.numpy_index_mult
                cond_2 = True
            except KeyError:
                if texture is None:
                    index_h, index_k, index_l, mult = crystal.calc_hkl(
                        sthovl_min, sthovl_max)
                else:
                    index_h, index_k, index_l, mult = \
                        crystal.calc_hkl_in_range(sthovl_min, sthovl_max)
                peak = TOFPeakL(loop_name=phase_label)
                peak.numpy_index_h = numpy.array(index_h, dtype=int)
                peak.numpy_index_k = numpy.array(index_k, dtype=int)
                peak.numpy_index_l = numpy.array(index_l, dtype=int)
                peak.numpy_index_mult = numpy.array(mult, dtype=int)
                d_internal_val[f"peak_{crystal.data_name:}"] = peak
                cond_2 = False

            cond_1 = not(crystal.is_variables())
            if cond_1 & cond_2:
                np_iint_u = peak.numpy_intensity_plus
                np_iint_d = peak.numpy_intensity_minus
            else:
                np_iint_u, np_iint_d = self.calc_iint(
                    index_h, index_k, index_l, crystal,
                    flag_internal=flag_internal)
                peak.numpy_intensity_plus = np_iint_u
                peak.numpy_intensity_minus = np_iint_d

            cell = crystal.cell
            sthovl_hkl = cell.calc_sthovl(index_h, index_k, index_l)
            d_hkl = 0.5/sthovl_hkl
            time_hkl = tof_parameters.calc_time_by_d(d_hkl)
            

            profile_2d = self.calc_shape_profile(
                d, d_hkl, phase_igsize=phase_igsize)

            peak.numpy_time = time_hkl

            tth_rad = tof_parameters.ttheta_bank * numpy.pi/180.
            wavelength = 2.*d_hkl*numpy.sin(0.5*tth_rad)
            wavelength_4 = wavelength**4
            
            iint_u_0 = np_iint_u * mult * wavelength_4
            iint_d_0 = np_iint_d * mult * wavelength_4

            if tof_parameters.is_attribute("extinction"):
                tof_extinction = tof_parameters.extinction
                iint_u_0 = (1. - tof_extinction*iint_u_0)*mult*wavelength_4
                iint_d_0 = (1. - tof_extinction*iint_d_0)*mult*wavelength_4

            np_iint_u_2d = numpy.meshgrid(time, iint_u_0, indexing="ij")[1]
            np_iint_d_2d = numpy.meshgrid(time, iint_d_0, indexing="ij")[1]

            # texture
            if texture is not None:
                cond_2 = ((not(texture.is_variables())) &
                          (not(cell.is_variables())))
                if cond_2:
                    try:
                        texture_2d = d_internal_val[
                            f"texture_2d_{crystal.data_name:}"]
                        cond_1 = False
                    except KeyError:
                        cond_1 = True
                if cond_1:
                    cos_alpha_ax = calc_cos_ang(cell, h_ax, k_ax, l_ax,
                                                index_h, index_k, index_l)
                    c_help = 1.-cos_alpha_ax**2
                    c_help[c_help < 0.] = 0.
                    sin_alpha_ax = numpy.sqrt(c_help)
                    cos_alpha_ax_2d = numpy.meshgrid(time, cos_alpha_ax,
                                                     indexing="ij")[1]
                    sin_alpha_ax_2d = numpy.meshgrid(time, sin_alpha_ax,
                                                     indexing="ij")[1]
                    # cos_alpha_2d = cos_alpha_ax_2d*cos_alpha_ang_2d + \
                    #                sin_alpha_ax_2d*sin_alpha_ang_2d
                    # cos_alpha_ang_2d, sin_alpha_ang_2d
                    cos_alpha_2d = cos_alpha_ax_2d*1.+sin_alpha_ax_2d*0.
                    texture_2d = g_2 + (1. - g_2) * (1./g_1 + (g_1**2 - 1./g_1)
                                                     * cos_alpha_2d**2)**(-1.5)
                    d_internal_val[f"texture_2d_{crystal.data_name:}"] = \
                        texture_2d

                profile_2d = profile_2d*texture_2d

            res_u_2d = profile_2d*np_iint_u_2d
            res_d_2d = profile_2d*np_iint_d_2d

            # 0.5 to have the same meaning for scale factor as in FullProf
            res_u_1d += 0.5*scale*res_u_2d.sum(axis=1)
            res_d_1d += 0.5*scale*res_d_2d.sum(axis=1)

            if flag_internal:
                peak.numpy_to_items()

        tof_proc.numpy_d_spacing = d
        tof_proc.numpy_intensity_plus_net = res_u_1d
        tof_proc.numpy_intensity_minus_net = res_d_1d
        tof_proc.numpy_intensity_net = res_u_1d+res_d_1d
        tof_proc.numpy_intensity_diff_total = res_u_1d-res_d_1d
        if flag_polarized:
            tof_proc.numpy_intensity_plus_total = res_u_1d+int_bkgd
            tof_proc.numpy_intensity_minus_total = res_d_1d+int_bkgd
            tof_proc.numpy_intensity_total = res_u_1d+res_d_1d+int_bkgd+int_bkgd
        else:
            tof_proc.numpy_intensity_plus_total = res_u_1d+0.5*int_bkgd
            tof_proc.numpy_intensity_minus_total = res_d_1d+0.5*int_bkgd
            tof_proc.numpy_intensity_total = res_u_1d+res_d_1d+int_bkgd

        if flag_internal:
            tof_proc.numpy_to_items()
            l_calc_objs = []
            for crystal in l_crystal:
                try:
                    obj = d_internal_val[f"refln_{crystal.data_name}"]
                    l_calc_objs.append(obj)
                except KeyError:
                    pass
                try:
                    obj = d_internal_val[
                        f"refln_susceptibility_{crystal.data_name}"]
                    l_calc_objs.append(obj)
                except KeyError:
                    pass
                try:
                    obj = d_internal_val[f"peak_{crystal.data_name}"]
                    l_calc_objs.append(obj)
                except KeyError:
                    pass
            l_calc_objs.append(tof_proc)
            self.add_items(l_calc_objs)
        return tof_proc

    def calc_chi_sq(self, l_crystal, flag_internal=True):
        """
        Calculate chi square.

        Arguments
        ---------
            - l_crystal: a list of Crystal objects of cryspy library
            - flag_internal: a flag to calculate or to use internal objects.
                   It should be True if user call the function.
                   It's True by default.

        Output
        ------
            - chi_sq_val: chi square of flip ratio
              (Sum_i ((y_e_i - y_m_i) / sigma_i)**2)
            - n: number of measured reflections
        """
        tof_meas = self.tof_meas
        flag_polarized = tof_meas.is_polarized()

        np_time = tof_meas.numpy_time
        if flag_polarized:
            int_u_exp = tof_meas.numpy_intensity_plus
            sint_u_exp = tof_meas.numpy_intensity_plus_sigma
            int_d_exp = tof_meas.numpy_intensity_minus
            sint_d_exp = tof_meas.numpy_intensity_minus_sigma
        else:
            int_exp = tof_meas.numpy_intensity
            sint_exp = tof_meas.numpy_intensity_sigma

        cond_in = numpy.ones(np_time.shape, dtype=bool)
        try:
            range_ = self.range
            time_min = numpy.array(range_.time_min, dtype=float)
            time_max = numpy.array(range_.time_max, dtype=float)

            cond_in = numpy.logical_and(cond_in, np_time >= time_min)
            cond_in = numpy.logical_and(cond_in, np_time <= time_max)
        except AttributeError:
            pass

        np_time_in = np_time[cond_in]
        if flag_polarized:
            int_u_exp_in = int_u_exp[cond_in]
            sint_u_exp_in = sint_u_exp[cond_in]
            int_d_exp_in = int_d_exp[cond_in]
            sint_d_exp_in = sint_d_exp[cond_in]
        else:
            int_exp_in = int_exp[cond_in]
            sint_exp_in = sint_exp[cond_in]

        tof_proc = self.calc_profile(
            np_time_in, l_crystal, flag_internal=flag_internal,
            flag_polarized=flag_polarized)

        if flag_polarized:
            tof_proc.numpy_intensity_plus = int_u_exp_in
            tof_proc.numpy_intensity_plus_sigma = sint_u_exp_in
            tof_proc.numpy_intensity_minus = int_d_exp_in
            tof_proc.numpy_intensity_minus_sigma = sint_d_exp_in
            tof_proc.numpy_intensity = int_u_exp_in+int_d_exp_in
            tof_proc.numpy_intensity_sigma = numpy.sqrt(
                numpy.square(sint_u_exp_in) + numpy.square(sint_d_exp_in))
        else:
            tof_proc.numpy_intensity = int_exp_in
            tof_proc.numpy_intensity_sigma = sint_exp_in

        int_u_mod = tof_proc.numpy_intensity_plus_total
        int_d_mod = tof_proc.numpy_intensity_minus_total

        if flag_polarized:
            sint_sum_exp_in = (sint_u_exp_in**2 + sint_d_exp_in**2)**0.5
            chi_sq_u = ((int_u_mod-int_u_exp_in)/sint_u_exp_in)**2
            chi_sq_d = ((int_d_mod-int_d_exp_in)/sint_d_exp_in)**2
            chi_sq_sum = ((int_u_mod+int_d_mod-int_u_exp_in-int_d_exp_in) /
                          sint_sum_exp_in)**2
            chi_sq_dif = ((int_u_mod-int_d_mod-int_u_exp_in+int_d_exp_in) /
                          sint_sum_exp_in)**2

            cond_u = numpy.logical_not(numpy.isnan(chi_sq_u))
            cond_d = numpy.logical_not(numpy.isnan(chi_sq_d))
            cond_sum = numpy.logical_not(numpy.isnan(chi_sq_sum))
            cond_dif = numpy.logical_not(numpy.isnan(chi_sq_dif))
        else:
            chi_sq_sum = ((int_u_mod+int_d_mod-int_exp_in)/sint_exp_in)**2
            cond_sum = numpy.logical_not(numpy.isnan(chi_sq_sum))

        # exclude region
        try:
            exclude = self.exclude
            l_excl_time_min = exclude.time_low
            l_excl_time_max = exclude.time_high
            if flag_polarized:
                for excl_time_min, excl_time_max in zip(l_excl_time_min,
                                                        l_excl_time_max):
                    cond_1 = numpy.logical_or(np_time_in < 1.*excl_time_min,
                                              np_time_in > 1.*excl_time_max)
                    cond_u = numpy.logical_and(cond_u, cond_1)
                    cond_d = numpy.logical_and(cond_d, cond_1)
                    cond_sum = numpy.logical_and(cond_sum, cond_1)
            else:
                for excl_time_min, excl_time_max in zip(l_excl_time_min,
                                                        l_excl_time_max):
                    cond_1 = numpy.logical_or(np_time_in < 1.*excl_time_min,
                                              np_time_in > 1.*excl_time_max)
                    cond_sum = numpy.logical_and(cond_sum, cond_1)
        except AttributeError:
            pass

        tof_proc.numpy_excluded = numpy.logical_not(cond_sum)
        chi_sq_sum_val = (chi_sq_sum[cond_sum]).sum()
        n_sum = cond_sum.sum()

        if flag_polarized:

            chi_sq_u_val = (chi_sq_u[cond_u]).sum()
            n_u = cond_u.sum()

            chi_sq_d_val = (chi_sq_d[cond_d]).sum()
            n_d = cond_d.sum()

            chi_sq_dif_val = (chi_sq_dif[cond_dif]).sum()
            n_dif = cond_dif.sum()

            chi2 = self.chi2

            flag_u = chi2.up
            flag_d = chi2.down
            flag_sum = chi2.sum
            flag_dif = chi2.diff

            chi_sq_val = (int(flag_u)*chi_sq_u_val + int(flag_d)*chi_sq_d_val +
                          int(flag_sum)*chi_sq_sum_val +
                          int(flag_dif)*chi_sq_dif_val)
            n = (int(flag_u)*n_u + int(flag_d)*n_d + int(flag_sum)*n_sum +
                 int(flag_dif)*n_dif)
        else:
            chi_sq_val = chi_sq_sum_val
            n = n_sum

        if flag_internal:
            refine_ls = RefineLs(number_reflns=n,
                                 goodness_of_fit_all=chi_sq_val/float(n),
                                 weighting_scheme="sigma")
            self.refine_ls = refine_ls
            tof_proc.numpy_to_items()
        return chi_sq_val, n

    # def simulation(self, l_crystal, ttheta_start: float = 4.,
    #                ttheta_end: float = 120., ttheta_step: float = 0.1,
    #                flag_polarized: bool = True) -> NoReturn:
    #     """Simulate."""
    #     n_points = int(round((ttheta_end-ttheta_start)/ttheta_step))
    #     tth_in = numpy.linspace(ttheta_start, ttheta_end, n_points)
    #     if isinstance(l_crystal, Crystal):
    #         l_crystal = [l_crystal]
    #     self.calc_profile(tth_in, l_crystal, l_peak_in=None, l_refln_in=None,
    #                       l_refln_susceptibility_in=None, l_dd_in=None,
    #                       flag_internal=True, flag_polarized=flag_polarized)
    #     return

    def calc_iint(self, index_h, index_k, index_l, crystal: Crystal,
                  flag_internal: bool = True):
        """Calculate the integrated intensity for h, k, l reflections.

        Arguments
        ---------
            - h, k, l: 1D numpy array of Miller indexes, dtype = int32
            - l_crystal: a list of Crystal objects of cryspy library
            - flag_internal: a flag to calculate or to use internal objects.
                   It should be True if user call the function.
                   It's True by default.

        Output
        ------
            - iint_u: 1D numpy array of integrated intensity up, dtype = float
            - iint_d: 1D numpy array of integrated intensity up, dtype = float
            - refln: ReflnL object of cryspy library (nuclear structure factor)
            - refln_s: ReflnSusceptibilityL object of cryspy library
              (nuclear structure factor)
        """
        index_hkl = numpy.stack([index_h, index_k, index_l], axis=0)

        if flag_internal:
            try:
                d_internal_val = self.d_internal_val
            except AttributeError:
                d_internal_val = {}
                self.d_internal_val = d_internal_val
        else:
            d_internal_val = {}
            self.d_internal_val = d_internal_val

        tof_parameters = self.tof_parameters
        try:
            field = tof_parameters.field
        except AttributeError:
            field = 0.

        try:
            diffrn_radiation = self.diffrn_radiation
            p_u = float(diffrn_radiation.polarization)
            p_d = (2.*float(diffrn_radiation.efficiency)-1)*p_u
        except AttributeError:
            p_u = 0.0
            p_d = 0.0

        try:
            if (not(flag_internal) | crystal.is_variables()):
                raise KeyError
            refln = d_internal_val[f"refln_{crystal.data_name:}"]
        except KeyError:
            refln = crystal.calc_refln(index_h, index_k, index_l,
                                       flag_internal=flag_internal)
            d_internal_val[f"refln_{crystal.data_name:}"] = refln
        f_nucl = refln.numpy_f_calc
        f_nucl_sq = abs(f_nucl*f_nucl.conjugate())

        if isinstance(crystal, Crystal):
            try:
                if (not(flag_internal) | crystal.is_variables()):
                    raise KeyError
                refln_s = d_internal_val[
                    f"refln_susceptibility_{crystal.data_name:}"]
            except KeyError:
                refln_s = crystal.calc_refln_susceptibility(
                    index_h, index_k, index_l, flag_internal=flag_internal)
                d_internal_val[f"refln_susceptibility_{crystal.data_name:}"] \
                    = refln_s

            sft_11 = refln_s.numpy_chi_11_calc
            sft_12 = refln_s.numpy_chi_12_calc
            sft_13 = refln_s.numpy_chi_13_calc
            sft_21 = refln_s.numpy_chi_21_calc
            sft_22 = refln_s.numpy_chi_22_calc
            sft_23 = refln_s.numpy_chi_23_calc
            sft_31 = refln_s.numpy_chi_31_calc
            sft_32 = refln_s.numpy_chi_32_calc
            sft_33 = refln_s.numpy_chi_33_calc


            _11, _12 = field*sft_11, field*sft_12
            _21, _13 = field*sft_21, field*sft_13
            _22, _23 = field*sft_22, field*sft_23
            _31, _32 = field*sft_31, field*sft_32
            _33 = field*sft_33
            _ij = numpy.stack([_11, _12, _13, _21, _22, _23, _31, _32, _33], axis=0)
            cell = crystal.cell

            unit_cell_parameters = numpy.array([
                cell.length_a, cell.length_b, cell.length_c,
                cell.angle_alpha*numpy.pi/180.,
                cell.angle_beta*numpy.pi/180.,
                cell.angle_gamma*numpy.pi/180.], dtype=float)
            
            t_ij = calc_matrix_t(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=False)[0]
            #SIGMA = T CHI T^-1 = T chi T^T (because T is rotation matrix, therefore T^-1 = T^T)
            th_ij = calc_m1_m2_m1t(t_ij, _ij)[0]
            th_11, th_12, th_22 = th_ij[0], th_ij[1], th_ij[4]

            # fm_p_sq = (field**2)*abs(0.5*(th_11*th_11.conjugate()+
            #            th_22*th_22.conjugate())+th_12*th_12.conjugate())
            # fm_p_field = field*0.5*(th_11+th_22)
            fm_p_sq = abs(0.5 * (th_11 * th_11.conjugate() +
                                 th_22 * th_22.conjugate()) +
                          th_12 * th_12.conjugate())
            fm_p_field = 0.5*(th_11 + th_22)
            cross = 2.*(f_nucl.real*fm_p_field.real +
                        f_nucl.imag*fm_p_field.imag)

            iint_u = f_nucl_sq + fm_p_sq + p_u*cross
            iint_d = f_nucl_sq + fm_p_sq - p_d*cross

        elif isinstance(crystal, MagCrystal):
            try:
                if (not(flag_internal) | crystal.is_variables()):
                    raise KeyError
                f_mag_perp = d_internal_val[f"f_mag_perp_{crystal.data_name:}"]

            except KeyError:
                f_mag_perp = crystal.calc_f_mag_perp(index_h, index_k, index_l)
                d_internal_val[f"f_mag_perp_{crystal.data_name:}"] = f_mag_perp
            f_mag_perp_sq = abs(f_mag_perp*f_mag_perp.conjugate()).sum(axis=0)
            iint_u = f_nucl_sq + f_mag_perp_sq
            iint_d = f_nucl_sq + f_mag_perp_sq

        return iint_u, iint_d

    def _gauss_pd(self, time_2d):
        """One dimensional gauss powder diffraction."""
        ag, bg = self.ag, self.bg
        val_1 = bg*time_2d**2
        val_2 = numpy.where(val_1 < 5., numpy.exp(-val_1), 0.)
        self.gauss_pd = ag*val_2

    def _lor_pd(self, time_2d):
        """One dimensional lorentz powder diffraction."""
        al, bl = self.al, self.bl
        self.lor_pd = al*1./(1.+bl*time_2d**2)

    def calc_shape_profile(self, d, d_hkl, phase_igsize: float = 0.):
        """
        Calculate shape profile.

        Calculate profile in the range of time for reflections placed on
        time_hkl with i_g parameter by default equal to zero

        d, d_hkl in angstrems
        """
        tof_parameters = self.tof_parameters
        tof_profile = self.tof_profile
        
        time = tof_parameters.calc_time_by_d(d)
        time_hkl = tof_parameters.calc_time_by_d(d_hkl)

        # FIXME: strain_g, size_g, size_l
        np_shape_2d = tof_profile.calc_peak_shape_function(
            d, time, time_hkl, size_g=phase_igsize, strain_g=0.,
            size_l=0., strain_l=0.)

        # Lorentz factor
        tth_rad = tof_parameters.ttheta_bank*numpy.pi/180.
        lorentz_factor = 1./(numpy.sin(tth_rad)*numpy.sin(0.5*tth_rad))

        profile_2d = np_shape_2d*lorentz_factor

        return profile_2d

    def params_to_cif(self, separator="_", flag: bool = False,
                      flag_minimal: bool = True) -> str:
        """Save parameters to cif format."""
        ls_out = []
        l_cls = (DiffrnRadiation, Chi2, Range, Extinction, Texture,
                 TOFIntensityIncident, TOFParameters, TOFProfile, PhaseL,
                 ExcludeL, TOFBackground)
        l_obj = [item for item in self.items if type(item) in l_cls]
        l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
        l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
        return "\n".join(ls_out)

    def data_to_cif(self, separator="_", flag: bool = False,
                    flag_minimal: bool = True) -> str:
        """Save data to cif format."""
        ls_out = []
        l_cls = (TOFMeasL, )
        l_obj = [item for item in self.items if type(item) in l_cls]
        l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
        l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
        return "\n".join(ls_out)

    def calc_to_cif(self, separator="_", flag=False, flag_minimal=True) -> str:
        """Save calculations to cif format."""
        ls_out = []
        l_cls = (RefineLs, ReflnL, ReflnSusceptibilityL, TOFPeakL, TOFProcL)
        l_obj = [item for item in self.items if type(item) in l_cls]
        l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
        l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
        return "\n".join(ls_out)

    def apply_constraints(self):
        """Apply constraints."""
        if self.pd_meas is not None:
            self.pd_meas.apply_constraints()

    def plots(self):
        if self.is_attribute("tof_proc"):
            tof_proc = self.tof_proc
            res = tof_proc.plot_sum_diff()
            return [res,]
        elif self.is_attribute("tof_meas"):
            return self.tof_meas.plots()

    def get_dictionary(self):
        self.form_object()
        dict_tof = {}

        phase, range_, tof_background = None, None, None
        exclude = None
        tof_meas, tof_parameters, tof_profile = None, None, None
        diffrn_radiation, texture = None, None
        setup = None

        l_obj = take_items_by_class(self, (PhaseL, ))
        if len(l_obj) > 0:
            phase = l_obj[0]

        l_obj = take_items_by_class(self, (Range, ))
        if len(l_obj) > 0:
            range_ = l_obj[0]

        l_obj = take_items_by_class(self, (TOFBackground, ))
        if len(l_obj) > 0:
            tof_background = l_obj[0]

        l_obj = take_items_by_class(self, (ExcludeL, ))
        if len(l_obj) > 0:
            exclude = l_obj[0]

        l_obj = take_items_by_class(self, (TOFMeasL, ))
        if len(l_obj) > 0:
            tof_meas = l_obj[0]

        l_obj = take_items_by_class(self, (TOFParameters, ))
        if len(l_obj) > 0:
            tof_parameters = l_obj[0]

        l_obj = take_items_by_class(self, (TOFProfile, ))
        if len(l_obj) > 0:
            tof_profile = l_obj[0]

        l_obj = take_items_by_class(self, (DiffrnRadiation, ))
        if len(l_obj) > 0:
            diffrn_radiation = l_obj[0]

        l_obj = take_items_by_class(self, (TextureL, ))
        if len(l_obj) > 0:
            texture = l_obj[0]

        l_obj = take_items_by_class(self, (Setup, ))
        if len(l_obj) > 0:
            setup = l_obj[0]

        dict_tof["name"] = self.data_name
        dict_tof["type_name"] = self.get_name()

        if phase is not None:
            dict_tof["phase_name"] = numpy.array(phase.label, dtype=str)

            if phase.is_attribute("igsize"):
                dict_tof["phase_ig"] = numpy.array(phase.igsize, dtype=float)
                dict_tof["flags_phase_ig"] = numpy.array(phase.igsize_refinement, dtype=bool)
            else:
                dict_tof["phase_ig"] = numpy.zeros((len(phase.items),), dtype=float)
                dict_tof["flags_phase_ig"] = numpy.zeros((len(phase.items),), dtype=bool)

            if phase.is_attribute("scale"):
                dict_tof["phase_scale"] = numpy.array(phase.scale, dtype=float)
                dict_tof["flags_phase_scale"] = numpy.array(phase.scale_refinement, dtype=bool)
            else:
                dict_tof["phase_scale"] = numpy.zeros((len(phase.items),), dtype=float)
                dict_tof["flags_phase_scale"] = numpy.zeros((len(phase.items),), dtype=bool)

        if tof_background is not None:
            dict_tof["background_time_max"] = tof_background.time_max
            dict_tof["background_coefficients"] = tof_background.get_coefficients()
            dict_tof["flags_background_coefficients"] = tof_background.get_flags_coefficients()
            

        if tof_meas is not None:
            time = numpy.array(tof_meas.time, dtype=float)

            time_min = range_.time_min
            time_max = range_.time_max
            dict_tof["time_min"] = time_min
            dict_tof["time_max"] = time_max


            flag_in = numpy.logical_and(
                time >= time_min,
                time <= time_max)
            
            dict_tof["time"] = time[flag_in]
            if tof_meas.is_attribute("intensity_plus"):
                int_plus = numpy.array(tof_meas.intensity_plus, dtype=float)[flag_in]
                s_int_plus = numpy.array(tof_meas.intensity_plus_sigma, dtype=float)[flag_in]
                int_minus = numpy.array(tof_meas.intensity_minus, dtype=float)[flag_in]
                s_int_minus = numpy.array(tof_meas.intensity_minus_sigma, dtype=float)[flag_in]
                dict_tof["signal_exp_plus"] = numpy.stack([int_plus, s_int_plus], axis=0)
                dict_tof["signal_exp_minus"] = numpy.stack([int_minus, s_int_minus], axis=0)
            else:
                int_sum = numpy.array(tof_meas.intensity, dtype=float)[flag_in]
                s_int_sum = numpy.array(tof_meas.intensity_sigma, dtype=float)[flag_in]
                dict_tof["signal_exp"] = numpy.stack([int_sum, s_int_sum], axis=0)

            time_in_range = time[flag_in]
            flag_exclude = numpy.zeros(time_in_range.shape, dtype=bool)
            if exclude is not None:
                for item_e in exclude.items:
                    flag_in_1 = numpy.logical_and(
                        time_in_range >= item_e.time_min,
                        time_in_range <= item_e.time_max)
                    flag_exclude = numpy.logical_or(flag_exclude, flag_in_1)
            dict_tof["excluded_points"] = flag_exclude


        if tof_parameters is not None:
            dict_tof["ttheta_bank"] = tof_parameters.ttheta_bank * numpy.pi/180
            dict_tof["neutron_type"] = tof_parameters.neutrons
            dict_tof["zero"] = tof_parameters.zero
            dict_tof["dtt1"] = tof_parameters.dtt1
            if tof_parameters.is_attribute("dtt2"):
                dict_tof["dtt2"] = tof_parameters.dtt2
            if tof_parameters.is_attribute("zerot"):
                dict_tof["zerot"] = tof_parameters.zerot
            if tof_parameters.is_attribute("dtt1t"):
                dict_tof["dtt1t"] = tof_parameters.dtt1t
            if tof_parameters.is_attribute("dtt2t"):
                dict_tof["dtt2t"] = tof_parameters.dtt2t


        if tof_profile is not None:
            l_alpha = []
            l_alpha_refinement = []
            for numb in range(2):
                if tof_profile.is_attribute(f"alpha{numb:}"):
                    l_alpha.append(getattr(tof_profile, f"alpha{numb:}"))
                    l_alpha_refinement.append(getattr(tof_profile, f"alpha{numb:}_refinement"))
                else:
                    break

            l_beta = []
            l_beta_refinement = []
            for numb in range(2):
                if tof_profile.is_attribute(f"beta{numb:}"):
                    l_beta.append(getattr(tof_profile, f"beta{numb:}"))
                    l_beta_refinement.append(getattr(tof_profile, f"beta{numb:}_refinement"))
                else:
                    break

            l_gamma = []
            l_gamma_refinement = []
            for numb in range(2):
                if tof_profile.is_attribute(f"gamma{numb:}"):
                    l_gamma.append(getattr(tof_profile, f"gamma{numb:}"))
                    l_gamma_refinement.append(getattr(tof_profile, f"gamma{numb:}_refinement"))
                else:
                    break

            l_sigma = []
            l_sigma_refinement = []
            for numb in range(3):
                if tof_profile.is_attribute(f"sigma{numb:}"):
                    l_sigma.append(getattr(tof_profile, f"sigma{numb:}"))
                    l_sigma_refinement.append(getattr(tof_profile, f"sigma{numb:}_refinement"))
                else:
                    break

            if tof_profile.is_attribute("peak_shape"):
                peak_shape = tof_profile.peak_shape
            else:
                peak_shape = "pseudo-Voigt"

            dict_tof["profile_alphas"] = numpy.array(l_alpha, dtype=float)
            dict_tof["flags_profile_alphas"] = numpy.array(l_alpha_refinement, dtype=float)
            dict_tof["profile_betas"] = numpy.array(l_beta, dtype=float)
            dict_tof["flags_profile_betas"] = numpy.array(l_beta_refinement, dtype=float)
            dict_tof["profile_gammas"] = numpy.array(l_gamma, dtype=float)
            dict_tof["flags_profile_gammas"] = numpy.array(l_gamma_refinement, dtype=float)
            dict_tof["profile_sigmas"] = numpy.array(l_sigma, dtype=float)
            dict_tof["flags_profile_sigmas"] = numpy.array(l_sigma_refinement, dtype=float)
            dict_tof["profile_peak_shape"] = peak_shape

        if diffrn_radiation is not None:
            beam_polarization = diffrn_radiation.polarization
            flipper_efficiency = diffrn_radiation.efficiency
            dict_tof["beam_polarization"] = numpy.array([beam_polarization], dtype=float)
            dict_tof["flipper_efficiency"] = numpy.array([flipper_efficiency], dtype=float)
            dict_tof["flags_beam_polarization"] = numpy.array([diffrn_radiation.polarization_refinement], dtype=bool)
            dict_tof["flags_flipper_efficiency"] = numpy.array([diffrn_radiation.efficiency_refinement], dtype=bool)


        if texture is not None:
            dict_tof["texture_name"] = numpy.array(texture.label, dtype=str)
            dict_tof["texture_g1"] = numpy.array(texture.g_1, dtype=float)
            dict_tof["texture_g2"] = numpy.array(texture.g_2, dtype=float)
            dict_tof["texture_axis"] = numpy.array(
                [texture.h_ax, texture.k_ax, texture.l_ax], dtype=float)
            dict_tof["flags_texture_g1"] = numpy.array(texture.g_1_refinement, dtype=bool)
            dict_tof["flags_texture_g2"] = numpy.array(texture.g_2_refinement, dtype=bool)
            dict_tof["flags_texture_axis"] = numpy.array(
                [texture.h_ax_refinement, 
                 texture.k_ax_refinement, 
                 texture.l_ax_refinement], dtype=bool)


        if setup is not None:
            dict_tof["magnetic_field"] = numpy.array([setup.field], dtype=float)

        return dict_tof


    def take_parameters_from_dictionary(self, ddict_diffrn, l_parameter_name: list=None, l_sigma: list=None):
        keys = ddict_diffrn.keys()
        if l_parameter_name is not None:
            parameter_label = [hh[0] for hh in l_parameter_name]
        else:
            parameter_label = []
        if "beam_polarization" in keys:
            self.diffrn_radiation.polarization = ddict_diffrn["beam_polarization"]

        if "flipper_efficiency" in keys:
            self.diffrn_radiation.efficiency = ddict_diffrn["flipper_efficiency"]

        if "phase_scale" in keys:
            hh = ddict_diffrn["phase_scale"]
            for i_item, item in enumerate(self.phase.items):
                item.scale = float(hh[i_item])

        if "phase_ig" in keys:
            hh = ddict_diffrn["phase_ig"]
            for i_item, item in enumerate(self.phase.items):
                item.igsize = float(hh[i_item])


        if len(parameter_label) > 0:
            for name, sigma in zip(l_parameter_name, l_sigma):
                if name[0] == "phase_scale":
                    self.phase.items[name[1][0]].scale_sigma = float(sigma)
                if name[0] == "phase_ig":
                    self.phase.items[name[1][0]].igsize_sigma = float(sigma)
                if name[0] == "beam_polarization":
                    self.diffrn_radiation.polarization_sigma = float(sigma)
                if name[0] == "flipper_efficiency":
                    self.diffrn_radiation.efficiency_sigma = float(sigma)
                if name[0] == "texture_g1":
                    self.texture.items[name[1][0]].g1_sigma = float(sigma)
                if name[0] == "texture_g2":
                    self.texture.items[name[1][0]].g2_sigma = float(sigma)
                if name[0] == "texture_axis":
                    ind_param, ind_a = name[1]
                    if ind_param == 0:
                        self.texture.items[ind_a].h_ax_sigma = float(sigma)
                    if ind_param == 1:
                        self.texture.items[ind_a].k_ax_sigma = float(sigma)
                    if ind_param == 2:
                        self.texture.items[ind_a].l_ax_sigma = float(sigma)

        if (("signal_plus" in keys) and ("signal_minus" in keys)):
            tof_proc = TOFProcL()
            tof_proc.numpy_time = numpy.round(ddict_diffrn["time"], decimals=5)
            tof_proc.numpy_intensity_plus_net = numpy.round(ddict_diffrn["signal_plus"], decimals=5)
            tof_proc.numpy_intensity_minus_net = numpy.round(ddict_diffrn["signal_minus"], decimals=5)
            if "signal_exp_plus" in keys:
                tof_proc.numpy_intensity_plus = numpy.round(ddict_diffrn["signal_exp_plus"][0], decimals=5)
                tof_proc.numpy_intensity_plus_sigma = numpy.round(ddict_diffrn["signal_exp_plus"][1], decimals=5)
                tof_proc.numpy_intensity_minus = numpy.round(ddict_diffrn["signal_exp_minus"][0], decimals=5)
                tof_proc.numpy_intensity_minus_sigma = numpy.round(ddict_diffrn["signal_exp_minus"][1], decimals=5)
            else:
                tof_proc.numpy_intensity = numpy.round(ddict_diffrn["signal_exp"][0], decimals=5)
                tof_proc.numpy_intensity_sigma = numpy.round(ddict_diffrn["signal_exp"][1], decimals=5)
            tof_proc.numpy_intensity_bkg_calc = numpy.round(ddict_diffrn["signal_background"], decimals=5)
            tof_proc.numpy_excluded = ddict_diffrn["excluded_points"]
            tof_proc.numpy_to_items()
            self.tof_proc = tof_proc

        l_tof_peak = []
        l_refln = []
        for item_phase in self.phase.items:
            s_label = f"dict_in_out_{item_phase.label:}"
            if s_label in keys:
                dict_crystal = ddict_diffrn[s_label]
                dict_crystal_keys = dict_crystal.keys()
                if (("index_hkl" in dict_crystal_keys) and  ("time_hkl" in dict_crystal_keys)):
                    index_hkl = dict_crystal["index_hkl"]
                    time_hkl = dict_crystal["time_hkl"]

                    tof_peak = TOFPeakL(loop_name = item_phase.label)
                    int_plus_max, int_minus_max = None, None
                    if "iint_plus_with_factors" in dict_crystal_keys:
                        int_plus_max = dict_crystal["iint_plus_with_factors"]
                    if "iint_minus_with_factors" in dict_crystal_keys:
                        int_minus_max = dict_crystal["iint_minus_with_factors"]

                    if "f_nucl":
                        refln = ReflnL(loop_name = item_phase.label)
                        refln.numpy_index_h = index_hkl[0]
                        refln.numpy_index_k = index_hkl[1]
                        refln.numpy_index_l = index_hkl[2]
                        refln.numpy_a_calc = dict_crystal["f_nucl"].real
                        refln.numpy_b_calc = dict_crystal["f_nucl"].imag
                        refln.numpy_to_items()
                        l_refln.append(refln)

                    if not((int_plus_max is None) and (int_minus_max is None)):
                        tof_peak.numpy_index_h = index_hkl[0]
                        tof_peak.numpy_index_k = index_hkl[1]
                        tof_peak.numpy_index_l = index_hkl[2]
                        tof_peak.numpy_time = numpy.round(time_hkl, decimals=3)
                        tof_peak.numpy_intensity_plus = int_plus_max
                        tof_peak.numpy_intensity_minus = int_minus_max
                        tof_peak.numpy_to_items()
                        l_tof_peak.append(tof_peak)

        if len(l_tof_peak) > 0:
            self.add_items(l_tof_peak)
        if len(l_refln) > 0:
            self.add_items(l_refln)

    def estimate_background(self):
        tof_background = self.tof_background
        tof_proc = self.tof_proc
        tof_proc.estimate_background(tof_background)




"""Description of Crystal class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"
from typing import NoReturn
import numpy
from warnings import warn


from cryspy.A_functions_base.symmetry_elements import calc_full_mag_elems

from cryspy.A_functions_base.function_3_mcif import calc_full_sym_elems, \
    calc_hkl_in_range

from cryspy.B_parent_classes.cl_3_data import DataN

from cryspy.C_item_loop_classes.cl_1_cell import Cell
from cryspy.C_item_loop_classes.cl_1_atom_site import AtomSiteL
from cryspy.C_item_loop_classes.cl_1_atom_type import AtomTypeL
from cryspy.C_item_loop_classes.cl_1_atom_site_aniso import \
    AtomSiteAnisoL
from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_atom_site_moment import AtomSiteMomentL
from cryspy.C_item_loop_classes.cl_1_space_group_symop_magn_centering import \
    SpaceGroupSymopMagnCenteringL
from cryspy.C_item_loop_classes.cl_2_space_group_symop_magn_operation import \
    SpaceGroupSymopMagnOperationL

from cryspy.C_item_loop_classes.cl_1_atom_type_scat import \
    AtomTypeScatL
from cryspy.C_item_loop_classes.cl_2_atom_site_scat import \
    AtomSiteScatL
from cryspy.C_item_loop_classes.cl_1_atom_local_axes import \
    AtomLocalAxesL
from cryspy.C_item_loop_classes.cl_1_atom_electron_configuration \
    import AtomElectronConfigurationL

from cryspy.D_functions_item_loop.function_1_calc_for_magcrystal import \
    calc_f_nucl, calc_f_mag
from cryspy.D_functions_item_loop.function_1_flip_ratio import \
    calc_fm_perp_for_fm_loc

from cryspy.A_functions_base.matrix_operations import calc_vector_product_v1_v2_v1
from cryspy.A_functions_base.unit_cell import calc_eq_ccs_by_unit_cell_parameters


class MagCrystal(DataN):
    """
    Crystal structure description with defined magnetic space group.

    Data items in the MAG_CRYSTAL category record details about
    magnetic crystal structure.

    """

    CLASSES_MANDATORY = (Cell, AtomSiteL, AtomSiteMomentL,
                         SpaceGroupSymopMagnOperationL)
    CLASSES_OPTIONAL = (
        AtomTypeL, AtomSiteAnisoL, SpaceGroupSymopMagnCenteringL,
        AtomSiteScatL, AtomTypeScatL, AtomLocalAxesL,
        AtomElectronConfigurationL)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "mag_crystal"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, data_name=None, **kwargs) -> NoReturn:
        super(MagCrystal, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["data_name"] = data_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self) -> NoReturn:
        """Redefined method of DataN."""
        self.apply_constraints()

    def apply_constraints(self) -> NoReturn:
        """
        Symmetry constraints on parameters.

        Returns
        -------
        NoReturn
            DESCRIPTION.

        """
        try:
            atom_site = self.atom_site
            atom_site_scat = self.atom_site_scat
            atom_site_scat.load_atom_type_scat_by_atom_site(atom_site)
        except AttributeError:
            pass

        # space_group = self.space_group
        # space_group_wyckoff = space_group.space_group_wyckoff

        # cell = self.cell
        # cell.type_cell = space_group.bravais_type
        # cell.it_coordinate_system_code = space_group.it_coordinate_system_code
        # cell.apply_constraints()

        # atom_site = self.atom_site
        # atom_site.apply_constraints(space_group_wyckoff)
        # atom_site_aniso = self.atom_site_aniso
        # if atom_site_aniso is not None:
        #     atom_site_aniso.apply_space_group_constraint(atom_site,
        #                                                  space_group)
        # atom_site_susceptibility = self.atom_site_susceptibility
        # if atom_site_susceptibility is not None:
        #     atom_site_susceptibility.apply_chi_iso_constraint(cell)
        #     atom_site_susceptibility.apply_space_group_constraint(atom_site,
        #                                                           space_group)

    def calc_b_iso_beta(self):
        """
        Calculate b_iso and beta_ij based on atom_site and atom_sites.

        For each atom defined in atom_site.
        """
        pass

    def calc_f_nucl(self, index_hkl):
        """
        Calculate nuclear structure factor.

        Arguments
        ---------
            index_h, index_k, index_l: 1D numpy array of Miller indexes

        Output
        ------
            f_nucl: 1D numpy array of Nuclear structure factor

        Example
        -------
            >>> import numpy as np
            >>> h, k, l = np.array([1,2],dtype=int), np.array([1,0],dtype=int),
                          np.array([1,0],dtype=int)
            >>> f_nucl = crystal.calc_f_nucl(h, k, l)

        """
        cell = self.cell
        atom_site = self.atom_site
        try:
            atom_site_aniso = self.atom_site_aniso
        except AttributeError:
            atom_site_aniso = None

        space_group_symop_magn_operation = \
            self.space_group_symop_magn_operation
        space_group_symop_magn_centering = \
            self.space_group_symop_magn_centering

        f_hkl_as, dder = calc_f_nucl(
            index_hkl,
            space_group_symop_magn_operation,
            space_group_symop_magn_centering,
            cell, atom_site, atom_site_aniso, flag_derivatives=False)
        return f_hkl_as

    def calc_refln(self, index_hkl, flag_internal=True):
        """
        Calculate Refln cryspy object where nuclear structure factor is stored.

        Arguments
        ---------
            h, k, l: 1D numpy array of Miller indexes
            flag_internal: a flag to calculate or to use internal objects.
                           It should be True if user call the function.
                           It's True by default.

        Output
        ------
            refln: object cryspy.Refln

        """
        f_nucl = self.calc_f_nucl(index_hkl)
        res = ReflnL()
        res.numpy_index_h = index_hkl[0]
        res.numpy_index_k = index_hkl[1]
        res.numpy_index_l = index_hkl[2]
        res.numpy_f_calc = f_nucl
        if flag_internal:
            res.numpy_to_items()
        return res

    def calc_f_mag(self, index_hkl):
        """
        Calculate magnetic structure factor.

        Keyword Arguments:
        -----------------
            index_h, index_k, index_l: 1D numpy array of Miller indexes

        Output:
        -------
            f_mag: 2D numpy array of Nuclear structure factor

        Example:
        -------
            >>> import numpy as np
            >>> h, k, l = np.array([1,2],dtype=int), np.array([1,0],dtype=int),
                          np.array([1,0],dtype=int)
            >>> f_nucl = crystal.calc_f_nucl(h, k, l)

        """
        cell = self.cell
        atom_site = self.atom_site
        try:
            atom_site_aniso = self.atom_site_aniso
        except AttributeError:
            atom_site_aniso = None

        atom_site_scat = self.atom_site_scat
        atom_site_moment = self.atom_site_moment

        space_group_symop_magn_operation = \
            self.space_group_symop_magn_operation
        space_group_symop_magn_centering = \
            self.space_group_symop_magn_centering

        flag_only_orbital = False
        flag_derivatives = False
        f_mag, dder = calc_f_mag(
            index_hkl, space_group_symop_magn_operation,
            space_group_symop_magn_centering, cell, atom_site, atom_site_aniso,
            atom_site_scat, atom_site_moment,
            flag_derivatives=flag_derivatives,
            flag_only_orbital=flag_only_orbital)
        return f_mag

    def calc_f_mag_perp(self, index_hkl):
        """
        Calculate perpendicular component of magnetic structure factor.

        Keyword Arguments:
        -----------------
            index_h, index_k, index_l: 1D numpy array of Miller indexes

        Output:
        -------
            f_mag: 2D numpy array of Nuclear structure factor

        Example:
        -------
            >>> import numpy as np
            >>> h, k, l = np.array([1,2],dtype=int), np.array([1,0],dtype=int),
                          np.array([1,0],dtype=int)
            >>> f_nucl = crystal.calc_f_nucl(h, k, l)

        """
        f_mag = self.calc_f_mag(index_hkl)

        cell = self.cell
        unit_cell_parameters = cell.get_unit_cell_parameters()

        flag_unit_cell_parameters = False
        eq_ccs, dder_eq_ccs = calc_eq_ccs_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)
        f_mag_perp, dder_f_mag_perp = calc_vector_product_v1_v2_v1(eq_ccs, f_mag)
        return f_mag_perp

    def calc_hkl(self, sthovl_min: float, sthovl_max: float):
        """
        Give hkl taking symmetry into account.

        Parameters
        ----------
        sthol_min : float
            DESCRIPTION.
        sthovl_max : float
            DESCRIPTION.

        Returns
        -------
        res : TYPE
            DESCRIPTION.

        """
        cell = self.cell
        rad = numpy.pi / 180.
        unit_cell_parameters = (
            cell.length_a, cell.length_b, cell.length_c,
            cell.angle_alpha*rad, cell.angle_beta*rad, cell.angle_gamma*rad)

        se_sgsmo = self.space_group_symop_magn_operation.get_sym_elems()
        se_sgsmc = self.space_group_symop_magn_centering.get_sym_elems()

        elem_symm_fs = calc_full_mag_elems(se_sgsmo, se_sgsmc)

        ind_hkl_mult = calc_hkl_in_range(
            elem_symm_fs[:13], unit_cell_parameters, sthovl_min, sthovl_max)

        return ind_hkl_mult


    def calc_hkl_in_range(self, sthovl_min: float, sthovl_max: float):
        """
        Give hkl without taking symmetry into account.

        Parameters
        ----------
        sthol_min : float
            DESCRIPTION.
        sthovl_max : float
            DESCRIPTION.

        Returns
        -------
        res : TYPE
            DESCRIPTION.

        """
        cell = self.cell
        rad = numpy.pi / 180.
        unit_cell_parameters = (
            cell.length_a, cell.length_b, cell.length_c,
            cell.angle_alpha*rad, cell.angle_beta*rad, cell.angle_gamma*rad)

        full_sym_elems = numpy.array([
            [0], [0], [0], [1], [1], [0], [0], [0], [1], [0], [0], [0], [1],
            [1], [0], [0], [0], [1], [0], [0], [0], [1]], dtype=int)

        ind_hkl_mult = calc_hkl_in_range(
            full_sym_elems, unit_cell_parameters, sthovl_min, sthovl_max)
        return ind_hkl_mult


# s_cont = """
#   data_Fe3O4
#   _cell_angle_alpha 90.0
#   _cell_angle_beta 90.0
#   _cell_angle_gamma 90.0
#   _cell_length_a 8.56212()
#   _cell_length_b 8.56212
#   _cell_length_c 8.56212
#   _space_group_it_coordinate_system_code 2
#   _space_group_IT_number    227

#   loop_
#   _atom_site_adp_type
#   _atom_site_B_iso_or_equiv
#   _atom_site_fract_x
#   _atom_site_fract_y
#   _atom_site_fract_z
#   _atom_site_label
#   _atom_site_occupancy
#   _atom_site_type_symbol
#   Uani 0.0 0.125 0.125 0.125 Fe3A 1.0 Fe3+
#   Uani 0.0 0.5 0.5 0.5 Fe3B 1.0 Fe3+
#   Uiso 0.0 0.25521 0.25521 0.25521 O1 1.0 O2-

#   loop_
#   _atom_type_scat_length_neutron
#   _atom_type_symbol
#     0.945 Fe3+
#   0.5803 O2-

#   loop_
#   _atom_site_aniso_U_11
#   _atom_site_aniso_U_12
#   _atom_site_aniso_U_13
#   _atom_site_aniso_U_22
#   _atom_site_aniso_U_23
#   _atom_site_aniso_U_33
#   _atom_site_aniso_label
#   0.0 0.0 0.0 0.0 0.0 0.0 Fe3A
#   0.0 0.0 0.0 0.0 0.0 0.0 Fe3B

#   loop_
#   _atom_site_scat_label
#   _atom_site_scat_lande
#   Fe3A 2.0
#   Fe3B 2.0

#   loop_
#   _atom_site_susceptibility_label
#   _atom_site_susceptibility_chi_type
#   _atom_site_susceptibility_chi_11
#   _atom_site_susceptibility_chi_12
#   _atom_site_susceptibility_chi_13
#   _atom_site_susceptibility_chi_22
#   _atom_site_susceptibility_chi_23
#   _atom_site_susceptibility_chi_33
#   Fe3A Cani -3.468(74) 0.0 0.0 -3.468 0.0 -3.468
#   Fe3B Cani 3.041      0.0 0.0  3.041(21) 0.0  3.041

# """

# obj = Crystal.from_cif(s_cont)
# for var_name in obj.get_variable_names():
#     print(var_name)
#     obj.set_variable_by_name(var_name, 7)
# print(obj.atom_site_susceptibility)
# obj.apply_constraints()
# print(obj.atom_site_susceptibility)
# print(obj.report_main_axes_of_magnetization_ellipsoids())
